
package sam.antlr.autolog ;

import org.antlr.v4.runtime.* ;
import org.antlr.v4.runtime.tree.* ;
//----------------------------------
import java.io.* ; 
import java.util.* ; 
import java.nio.charset.* ;    // read from command line
import sam.machine.lang.*  ;   // machine lang translation target

/**
  *  Compiler   
  *     Parsed AUTOLOG.g4 theory --> sam.machine.lang.Theory
  *  @version 8/28/2018   
  *  ANTLR4 generates AUTOLOGListener and AUTOLOGBaseListener.
  *  Reader extends and breathes life into the AUTOLOGBaseListener methods.
  *  Compiler translates input theory to sam.machine.lang.Theory  
   */
public class Compiler {  

   /**
     *  Static translate Colog theory via InputStream to sam.machine.lang components
     *  Some DETAILS still pending ...
      */
   public static Theory translate(InputStream is) throws Exception  { 
      //ANTLRInputStream input = new ANTLRInputStream(is) ;
      CharStream input = CharStreams.fromStream(is) ;  // ANTLR 4.7
      AUTOLOGLexer lexer = new AUTOLOGLexer(input);
      CommonTokenStream tokens = new CommonTokenStream(lexer);
      AUTOLOGParser parser = new AUTOLOGParser(tokens);
      ParseTree tree = parser.theory();  // input Autolog theory parsed
      // create a standard ANTLR parse tree walker
      ParseTreeWalker walker = new ParseTreeWalker();
      Reader reader = new Reader();
      //System.out.println("Got to walker") ;
      walker.walk(reader, tree);     // walk parse tree, generate theory 
      // theory built by the walk
      return reader.theory ;

   }

   /**  
     *  Test stub for development.  
     *  From the command line, show translation of a theory:
     *    >java -cp . sam.antlr.autolog.Compiler user
     *        p, q => r(X).
     *        r(Z) => goal.
     *      ^D    ...
     *
     *  Using JavApp, enter one rule 
     *       Lit => test.
     *  in "command line parameters" box to inspect Lit.
      */
   public static void main(String[] args) { 

      if (args[0].trim().equals("user")) { 
         InputStream is = System.in ; 
         try { 
            Theory theory = translate(is) ; 
            System.out.println(theory) ;  // toString() 
            //System.out.println(theory.toCode()) ;   
         } catch (Exception xx) {System.out.println(xx);} 
         return ;
      }
      // Otherwise, DEBUG Functor from command line box of JavApp 
      StringBuffer sb = new StringBuffer() ;
      // build from the args-line so can test with main
      for (int i = 0 ; i < args.length ; i++) sb.append(args[i]+" ") ;
      System.out.println("RAW text input: " +sb) ;      
      // read from command line
      InputStream is = new ByteArrayInputStream(sb.toString().getBytes(Charset.forName("UTF-8"))) ; 
      try { 
         Theory thy = translate(is) ; 
         System.out.println("Theory toString(): ") ;
         System.out.println(thy) ;
         Rule rule = (Rule)thy.elementAt(0) ; 
         Functor f = rule.antecedent.elementAt(0) ;
         System.out.println("First lit(eral) of antecedent: "+f) ;
         System.out.println("lit.arity= "+f.arity) ; 
         System.out.println("lit.name= "+f.name) ;
         System.out.println("lit.fnctr= "+f.fnctr+ " Var="+(f.fnctr instanceof Variable)) ;
         System.out.println("lit.type= "+f.type) ;         
         //System.out.println("WANT name= "+f.name.substring(1,f.name.length()-1)) ;

      } catch (Exception xx) {System.out.println(xx);} 
      // Want to be able to inspect any term
   }

} // Compiler


/*
✓RAW text input: P(a)(b)(X) => tst. 
Theory toString(): 
P(a)(b)(X) => tst.
First lit(eral) of antecedent: P(a)(b)(X)
lit.arity= 1
lit.name= ‹P(a)(b)›
lit.fnctr= P(a)(b) Var=false
lit.type= 0

✓RAW text input: (f+g)(d)=(f(a)+g(a)) => tst. 
Theory toString(): 
(f+g)(d)=(f(a)+g(a)) => tst.
First lit(eral) of antecedent: (f+g)(d)=(f(a)+g(a))
lit.arity= 2
lit.name= =
lit.fnctr= null Var=false
lit.type= 2


✓RAW text input: (f+(g+h))(a)=(f(a)+(g+h)(a)) => tst. 
Theory toString(): 
(f+(g+h))(a)=(f(a)+(g+h)(a)) => tst.
First lit(eral) of antecedent: (f+(g+h))(a)=(f(a)+(g+h)(a))
lit.arity= 2
lit.name= =
lit.fnctr= null Var=false
lit.type= 2

✓RAW text input: ((P(a)›(b))(c))(X) => tst. 
line 1:6 token recognition error at: '›'
Theory toString(): 
P(a)(b)(c)(X) => tst.
First lit(eral) of antecedent: P(a)(b)(c)(X)
lit.arity= 1
lit.name= ‹P(a)(b)(c)›
lit.fnctr= P(a)(b)(c) Var=false
lit.type= 0
*/
