package sam.gui.edit ;     
/*  notes  */
import java.awt.* ; 
import java.awt.datatransfer.* ; 
import java.awt.event.* ;      
import java.awt.print.* ; 
import java.util.* ;
import java.io.* ;
import javax.swing.* ; 
import javax.swing.border.* ; 
import javax.swing.event.* ; 
import javax.swing.undo.* ; 
import javax.swing.text.* ;
import java.net.* ; 
import java.util.zip.* ;
import javax.swing.colorchooser.* ;
import java.nio.charset.* ;    // read from command line

/**
  *  AutoLogEditor 
  *  john.r.fisher@protonmail.com
  *  @author &copy; John R. Fisher
  *  @version 7/5/2023
   */
public class AutoLogEditor extends JFrame {  

   /////
   // Use "compile" command to show and update autologGUI.
   /////
   protected sam.gui.search.AutoLog autologGUI ;  
   protected void createAutoLogGUI() { 
      autologGUI = new sam.gui.search.AutoLog(0) ; 
      autologGUI.editor = this ; // GUI knows editor
   } 

   private JToolBar toolbar ; 
   public static String version = " 7/5/2023" ;
   private Properties settings ; 
   private JTextPane ta ; 
   private JTextArea linenumbers; 
   private int number_of_lines ; 

   // preferred font
   private String fontName ;
   private int fontStyle ;
   private int fontSize ; 
   // preferred key accelerator 
   private int keyAccelerator ; 
   private JScrollBar vert ; 
   public boolean changed ; 
   
   private JMenuItem compile ;
//   private JMenuItem run ;
   private JButton compileButton ;
//   private JButton runButton ;
   /////
   private JButton circleButton ; // ∘
   private JButton ccircleButton ; // ∙
   private JButton crossButton ; // ⨯
   private JButton meetButton ; // ∧
   private JButton joinButton ; // ∨
   private JButton notButton ;  // ¬
   private JButton ifButton ;   // →
   private JButton fiButton ;   // ←
   private JButton iffButton ;  // ↔
   private JButton allButton ;  // ∀
   private JButton someButton ; // ∃ 
   private JButton lambdaButton ; // λ
   private JButton piButton ; // Π
   private JButton sigmaButton ; // Σ
   private JButton possibleButton ; // ◇
   private JButton necessaryButton ; // □
   private JButton topButton ; // ⊤ 
   private JButton bottomButton ; // ⊥ 
   private JButton intersectButton ;  // ∩
   private JButton unionButton ;  // ∪
   private JButton emptyButton ;  // ∅
   private JButton memberButton ;  // ∈
   private JRadioButtonMenuItem fourteen ;    // 3/1/2019 re format
   private JRadioButtonMenuItem twentyfour ; 

   // default colors, unless loaded from saved settings
   public static Color 
       backgroundColor = Color.white ,   
       selectionColor = Color.LIGHT_GRAY,     
       fontColor = Color.black, 
       commentColor = Color.GREEN.darker().darker().darker(),          
       reservedwordColor =  Color.BLUE ;    

   public static Color shadeColor = Color.lightGray ; //new Color(240,240,240) ;

   public final String[] reservedwords = {"true", "false", "=>", "goal", ":", "|"  } ;   

   public SimpleAttributeSet fontForeground,
                            commentForeground,
                            reservedWordForeground ;

   //  Keep track of undo/redo
   protected UndoAction undoAction ;
   protected RedoAction redoAction ;
   protected UndoManager undoManager = new UndoManager() ;

   // Keep track of source
   private File fileSpec ;  
   private JFileChooser fileChooser ; 
   private String classPath ; 
   private String classSpec ; 


   private JPopupMenu popedit ; 
   private JLabel news ; 
   private JavaConsole console ; 
   private boolean console_visible,  colorWidget_visible ;
   private PageFormat pageformat ;  // for printing, print preview

   // Line numbers and decorated text flags
   private boolean linenumbersFlag ; 
   private boolean decorateTextFlag ; 
   // Keep track of local insert or remove for decoration
   private Point redecorate ; // redocorate from x to y when necessary

   JPanel togetherPanel ;  // holds linenumbers and text area

   private final AutoLogEditor activatorFor = this ;  // callback to this if needed in activator

   public FileActivator fileActivator  ; // scope visible to all
   private CommandActivator commandActivator ; 
   private EditActivator editActivator ; 

   private FullColorWidget colorWidget ; 

   public AutoLogEditor(String title) {
      super(title) ; 
      autologGUI = null ; // not created
      fontForeground = new SimpleAttributeSet() ; 
      StyleConstants.setForeground(fontForeground,fontColor) ; 
      commentForeground = new SimpleAttributeSet() ; 
      StyleConstants.setForeground(commentForeground,commentColor) ; 
      reservedWordForeground = new SimpleAttributeSet() ; 
      StyleConstants.setForeground(reservedWordForeground,reservedwordColor) ; 

      // preferred font & key accelerator
      if (System.getProperty("os.name").indexOf("Mac") >= 0) {
         fontName = "monaco" ; //  My favorite on the Mac
         keyAccelerator = InputEvent.META_DOWN_MASK ; 
      }
      else {
         fontName = "monospaced" ;  // do better if you can 
         keyAccelerator = InputEvent.CTRL_DOWN_MASK ; 
      }
      fontStyle = Font.PLAIN ;
      // fontSize gotten from settings below ...

      Class clz = this.getClass() ; 
      settings = new Properties() ; 
      // otherwise use built-in's
      try { 
         settings.load(
           new FileInputStream(
            System.getProperty("user.home")+File.separator+"AutoLogEditor.settings" )  ) ; 
      }
      catch(Exception x) { // first time, no settings (not fool proof)
         System.out.println(x) ; 
         // built-in ...
         settings.put("width","500") ; 
         settings.put("height","555") ; 
         settings.put("x","400") ; 
         settings.put("y","100") ; 
         settings.put("fontsize","18") ;
         settings.put("backgroundcolor",
                               String.valueOf(backgroundColor.getRed())+
                                ","+String.valueOf(backgroundColor.getGreen())+
                                      ","+String.valueOf(backgroundColor.getBlue())) ; 
         settings.put("selectioncolor",
                               String.valueOf(selectionColor.getRed())+
                                ","+String.valueOf(selectionColor.getGreen())+
                                      ","+String.valueOf(selectionColor.getBlue())) ; 
      } 
      try  { //... to use properties to size and locate
         this.setSize(Integer.parseInt(settings.getProperty("width").trim()),
                           Integer.parseInt(settings.getProperty("height").trim())) ; 
         this.setLocation(Integer.parseInt(settings.getProperty("x").trim()),
                           Integer.parseInt(settings.getProperty("y").trim())) ; 
      } 
      catch(Exception x) { 
         this.setSize(500,500) ; 
         this.setLocation(400,100) ; 
      }

      fileActivator = new FileActivator() ; 
      commandActivator = new CommandActivator() ; 
      editActivator = new EditActivator() ; 

      // prevent closing unless requested 
      this.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE) ; 

      // Start where user is
      fileSpec = new File(System.getProperty("user.home")) ; 
      fileChooser = null ; // initialized when first needed


      // What kind of system are we running? Try to avoid using.
      String system = System.getProperty("os.name") ; 
      boolean windows = system.indexOf("Windows") >= 0 ; 

      // CREATE UNDO AND REDO ACTIONS
      undoAction = new UndoAction() ;
      undoAction.putValue(Action.ACCELERATOR_KEY,KeyStroke.getKeyStroke(KeyEvent.VK_U,keyAccelerator)) ;
      redoAction = new RedoAction() ; 
      redoAction.putValue(Action.ACCELERATOR_KEY,KeyStroke.getKeyStroke(KeyEvent.VK_R,keyAccelerator)) ;

      /*   N.B.
         undoAction and redoAction are the only Java Action(s) used.  Other controls use our own
         Activator(s).  Actions are difficult to use with iconic buttons or popup
         menus (accelerators show there). ????  
      */

      // Create a menubar for this
      JMenuBar mb = new JMenuBar() ; 
      this.setJMenuBar(mb)  ; 
      // Create a file menu
      JMenu fileMenu = new JMenu("<html><b><font color=\"BLACK\"> AutoLogEdit</font></b>") ; 
      mb.add(fileMenu) ;

      JMenuItem open = new JMenuItem("Open") ;
      open.addActionListener(fileActivator) ;
      open.setActionCommand("open") ;
      open.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O,keyAccelerator)) ; 
      fileMenu.add(open) ;
      JMenuItem save = new JMenuItem("Save") ;
      save.addActionListener(fileActivator) ;
      save.setActionCommand("save") ;
      save.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S,keyAccelerator)) ; 
      fileMenu.add(save) ;
      JMenuItem saveas = new JMenuItem("Save As") ;
      saveas.addActionListener(fileActivator) ;
      saveas.setActionCommand("saveas") ;
      saveas.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S,InputEvent.SHIFT_DOWN_MASK+keyAccelerator)) ; 
      fileMenu.add(saveas) ;
      ////
      fileMenu.addSeparator() ; 
      PrintActivator printActivator = new PrintActivator() ; 
      JMenuItem pagesetup = new JMenuItem("Page Setup") ; 
      pagesetup.addActionListener(printActivator) ;
      pagesetup.setActionCommand("pagesetup") ;
      fileMenu.add(pagesetup) ; 
      JMenuItem print = new JMenuItem("Print") ; 
      print.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_P,keyAccelerator)) ; 
      print.addActionListener(printActivator) ; 
      print.setActionCommand("print") ;
      fileMenu.add(print) ; 

      JMenu editMenu = new JMenu("Edit") ; 
      mb.add(editMenu) ; 
      JMenuItem selectAll = new JMenuItem("Select All") ; 
      selectAll.addActionListener(editActivator) ; 
      selectAll.setActionCommand("selectall") ; 
      selectAll.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_A,keyAccelerator)) ; 
      editMenu.add(selectAll) ; 
      editMenu.addSeparator() ; 
      JMenuItem copy = new JMenuItem("Copy") ; 
      copy.addActionListener(editActivator) ; 
      copy.setActionCommand("copy") ; 
      copy.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C,keyAccelerator)) ; 
      JMenuItem cut = new JMenuItem("Cut") ; 
      cut.addActionListener(editActivator) ; 
      cut.setActionCommand("cut") ; 
      cut.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X,keyAccelerator)) ; 
      JMenuItem paste = new JMenuItem("Paste") ; 
      paste.addActionListener(editActivator) ;  
      paste.setActionCommand("paste") ; 
      paste.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_V,keyAccelerator)) ; 
      editMenu.add(copy) ; 
      editMenu.add(cut) ; 
      editMenu.add(paste) ; 
      editMenu.addSeparator() ; 
      //editMenu.addSeparator() ; 
      JMenuItem undo = new JMenuItem("Undo") ; 
      undo.setAction(undoAction) ; // <---
      JMenuItem redo = new JMenuItem("Redo") ; 
      redo.setAction(redoAction) ; // <--
      editMenu.add(undo) ; 
      editMenu.add(redo) ; 
      editMenu.addSeparator() ; 
      JMenuItem find = new JMenuItem("Find") ; 
      find.addActionListener(editActivator) ; 
      find.setActionCommand("find") ; 
      find.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F,keyAccelerator)) ; 
      JMenuItem findagain = new JMenuItem("Find Again") ; 
      findagain.addActionListener(editActivator) ;      
      findagain.setActionCommand("findagain") ; 
      findagain.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_G,keyAccelerator)) ; 
      editMenu.add(find) ; 
      editMenu.add(findagain) ; 
      editMenu.addSeparator() ; 
      // 2016 auto indent selection
      JMenuItem indent = new JMenuItem("Indent selection") ; 
      indent.addActionListener(editActivator) ; 
      indent.setActionCommand("indent") ; 
      editMenu.add(indent) ; 
      editMenu.addSeparator() ; 
      // 2016 goto line
      JMenuItem goline = new JMenuItem("goto line") ; 
      goline.addActionListener(editActivator) ; 
      goline.setActionCommand("goline") ; 
      editMenu.add(goline) ; 
     
      // pop up with right click
      popedit = new JPopupMenu() ; 
      JMenuItem copy1 = new JMenuItem("Copy") ; 
      copy1.addActionListener(editActivator) ; 
      copy1.setActionCommand("copy") ; 
      JMenuItem cut1 = new JMenuItem("Cut") ; 
      cut1.addActionListener(editActivator) ; 
      cut1.setActionCommand("cut") ; 
      JMenuItem paste1 = new JMenuItem("Paste") ; 
      paste1.addActionListener(editActivator) ; 
      paste1.setActionCommand("paste") ; 
      popedit.add(copy1) ;  
      popedit.add(cut1) ; 
      popedit.add(paste1) ; 
      // see ta below

      OptionsActivator optionsActivator = new OptionsActivator() ; 
      JMenu optionsMenu = new JMenu("Options") ; 
      if (windows) optionsMenu.setMnemonic('O') ; 
      mb.add(optionsMenu) ; 
      // SHOW LINE NUMBERS OR NOT
      JCheckBoxMenuItem lineNumbersMenuItem = 
           new JCheckBoxMenuItem("Line Numbers",true) ; 
      lineNumbersMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_L,keyAccelerator)) ; 
      lineNumbersMenuItem.addActionListener(optionsActivator) ;
      lineNumbersMenuItem.setActionCommand("showlinenumbers") ;
      linenumbersFlag = true ; 
      optionsMenu.add(lineNumbersMenuItem) ; 
      // DECORATE TEXT OR NOT
      JCheckBoxMenuItem decorateTextMenuItem = 
           new JCheckBoxMenuItem("Color Decorate",false) ; 
      decorateTextMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_D,keyAccelerator)) ; 
      decorateTextMenuItem.addActionListener(optionsActivator) ;  
      decorateTextMenuItem.setActionCommand("decorate") ;  
      decorateTextFlag = false ; 
      optionsMenu.add(decorateTextMenuItem) ; 

      // SHOW TOOLBAR OR NOT
      JCheckBoxMenuItem toolbarTextMenuItem = 
           new JCheckBoxMenuItem("Toolbar",true) ; 
      toolbarTextMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_T,keyAccelerator)) ; 
      toolbarTextMenuItem.addActionListener(optionsActivator) ;  
      //action command is already "Toolbar" -- name on JCheckBoxMenuItem  
      optionsMenu.add(toolbarTextMenuItem) ; 

      // FONT SIZE
      JMenu fontSizeMenu = new JMenu("Font Size") ; 
      optionsMenu.add(fontSizeMenu) ; 
      ButtonGroup fontSizeGroup = new ButtonGroup();
      JRadioButtonMenuItem ten = new JRadioButtonMenuItem("10") ; 
      fontSizeGroup.add(ten) ; 
      ten.addActionListener(optionsActivator) ;
      ten.setActionCommand("fontsize10") ; 
      fontSizeMenu.add(ten) ;
      JRadioButtonMenuItem twelve = new JRadioButtonMenuItem("12") ; 
      fontSizeGroup.add(twelve) ; 
      twelve.addActionListener(optionsActivator) ;
      twelve.setActionCommand("fontsize12") ; 
      fontSizeMenu.add(twelve) ;
      fourteen = new JRadioButtonMenuItem("14") ; 
      fontSizeGroup.add(fourteen) ; 
      fourteen.addActionListener(optionsActivator) ;
      fourteen.setActionCommand("fontsize14") ; 
      fontSizeMenu.add(fourteen) ;
      JRadioButtonMenuItem sixteen = new JRadioButtonMenuItem("16") ; 
      fontSizeGroup.add(sixteen) ; 
      sixteen.addActionListener(optionsActivator) ;
      sixteen.setActionCommand("fontsize16") ; 
      fontSizeMenu.add(sixteen) ;
      JRadioButtonMenuItem eightteen = new JRadioButtonMenuItem("18") ; 
      fontSizeGroup.add(eightteen) ; 
      eightteen.addActionListener(optionsActivator) ;
      eightteen.setActionCommand("fontsize18") ; 
      fontSizeMenu.add(eightteen) ;
      JRadioButtonMenuItem twenty = new JRadioButtonMenuItem("20") ; 
      fontSizeGroup.add(twenty) ; 
      twenty.addActionListener(optionsActivator) ;
      twenty.setActionCommand("fontsize20") ; 
      fontSizeMenu.add(twenty) ;
      JRadioButtonMenuItem twentytwo = new JRadioButtonMenuItem("22") ; 
      fontSizeGroup.add(twentytwo) ; 
      twentytwo.addActionListener(optionsActivator) ;
      twentytwo.setActionCommand("fontsize22") ; 
      fontSizeMenu.add(twentytwo) ;
      twentyfour = new JRadioButtonMenuItem("24") ; 
      fontSizeGroup.add(twentyfour) ; 
      twentyfour.addActionListener(optionsActivator) ;
      twentyfour.setActionCommand("fontsize24") ; 
      fontSizeMenu.add(twentyfour) ;

      // COLOR CHOOSER
      JMenuItem colorMenuItem = new JMenuItem("Color Chooser") ; 
      colorMenuItem.addActionListener(optionsActivator) ; 
      colorMenuItem.setActionCommand("colors") ; 
      optionsMenu.add(colorMenuItem) ;

      optionsMenu.addSeparator() ; 
      JMenuItem formatMenu = new JMenu("format") ; 
      // Set format 14pt, !linenumbers
      JMenuItem tidy = new JMenuItem("tidy") ; 
      tidy.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_0,keyAccelerator)) ; // cmd-0
      tidy.addActionListener(optionsActivator) ;
      tidy.setActionCommand("tidy") ; 
      formatMenu.add(tidy) ;
      // Set format 24pt, linenumber
      JMenuItem big = new JMenuItem("big") ; 
      big.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_1,keyAccelerator)) ; // cmd-1
      big.addActionListener(optionsActivator) ;
      big.setActionCommand("big") ; 
      formatMenu.add(big) ;
      // Locate next line boundary
      JMenuItem boundary = new JMenuItem("boundary") ; 
      boundary.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_2,keyAccelerator)) ; // cmd-2
      boundary.addActionListener(optionsActivator) ;
      boundary.setActionCommand("boundary") ; 
      formatMenu.add(boundary) ;
      optionsMenu.add(formatMenu) ;
      optionsMenu.addSeparator() ; 

      // SAVE SETTINGS
      JMenuItem save_settings = new JMenuItem("Save Settings") ; 
      save_settings.addActionListener(optionsActivator) ; 
      save_settings.setActionCommand("save_settings") ; 
      optionsMenu.add(save_settings) ; 

      // CODE
      JMenu commandMenu = new JMenu("Code") ; 
      mb.add(commandMenu) ;
      compile = new JMenuItem("Compile") ;  
      compile.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_K,keyAccelerator)) ; 
      compile.addActionListener(commandActivator) ; 
      compile.setActionCommand("compile") ; 
      commandMenu.add(compile) ; 
      commandMenu.addSeparator() ; 
      /////
      JMenuItem info = new JMenuItem("<html><font color=\"RED\">?<font color=\"GREEN\">?<font color=\"BLUE\">?</font></font></font> code info") ;  
      info.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_I,keyAccelerator)) ; 
      info.addActionListener(commandActivator) ;     // >> CommandActivator
      info.setActionCommand("info") ; 
      commandMenu.add(info) ; 
      //  unicode symbol tables
      JMenuItem symbols = new JMenuItem("unicode symbols") ;  
      symbols.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_I,InputEvent.SHIFT_DOWN_MASK+keyAccelerator)) ; 
      symbols.addActionListener(commandActivator) ;  // >> CommandActivator
      symbols.setActionCommand("symbols") ;        
      commandMenu.add(symbols) ; 

      ta = new JTextPane() ;  // used to be JTextArea brfore decoration
      ta.setDragEnabled(true) ; 
      ta.setMargin(new Insets(0,5,0,5)) ; 
      ta.getDocument().addDocumentListener(new DocumentActivator()) ;  // Activator 
      ta.getDocument().addUndoableEditListener(new QUndoableEditListener()) ;  
      // right click on ta to get popedit
      ta.addMouseListener(editActivator) ;  

      // keep decoration current
      redecorate = new Point(-1,-1) ; 
      ta.addMouseMotionListener(new MouseMotionAdapter() { 
         public void mouseMoved(MouseEvent me) { 
            if (decorateTextFlag && redecorate.x >= 0)  { 
               redecorate(redecorate.x,redecorate.y) ; 
            }
            // reset for next ...
            redecorate.x = -1 ; 
            redecorate.y = -1 ; 
         }
      }) ; 

      linenumbers = new JTextArea() ; 
      linenumbers.setBackground(shadeColor) ; 
      linenumbers.setEditable(false) ; 
      // double click on line numbers to renumber
      linenumbers.addMouseListener(new MouseAdapter() { 
         public void mousePressed(MouseEvent me) { 
            if(me.getClickCount()>1)  { 
               int p = linenumbers.getCaretPosition() ; 
               renumberLines() ; 
               try { 
                  linenumbers.setCaretPosition(p) ; 
               } catch(Exception x) {} 
            }
         }
      }) ;            

      // minimal start message ... 
      news = new JLabel("<html>   <b>AutoLogEditor</b> is <i>ready</i>" + " </html> ") ; 
      togetherPanel = new JPanel() ; 

/////////////////////////////////////////////////////////////////
///// Toolbar ///////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////
      toolbar = new JToolBar() ; 
      toolbar.setBackground(shadeColor) ; 
      Class cls = this.getClass() ; 
      Font bfont = new Font("monospaced",Font.PLAIN,18) ;

      JButton saveButton = new JButton("Save") ; 
      saveButton.setFont(bfont) ;
      saveButton.createToolTip() ; 
      saveButton.setToolTipText("<html><font size=\"+2\">save to file") ; 
      saveButton.setBackground(shadeColor) ; 
      saveButton.addActionListener(fileActivator) ; 
      saveButton.setActionCommand("save") ;
      toolbar.add(saveButton) ; 
      //toolbar.add(new JToolBar.Separator()) ;       

      compileButton = new JButton("Compile") ; 
      compileButton.setFont(bfont) ;
      compileButton.createToolTip() ; 
      compileButton.setToolTipText("<html><font size=\"+2\">compile to AutoLogViewer") ; 
      compileButton.setBackground(shadeColor) ; 
      compileButton.addActionListener(commandActivator) ; 
      compileButton.setActionCommand("compile") ;
      toolbar.add(compileButton) ; 
      toolbar.add(new JLabel("   ")) ;       
      circleButton = new JButton("∘") ; 
      circleButton.setFont(bfont) ;  
      circleButton.createToolTip() ; 
      circleButton.setToolTipText("<html><font size=\"+2\">insert ∘ (composition) at cursor") ; 
      circleButton.setBackground(shadeColor) ; 
      circleButton.addActionListener(commandActivator) ; 
      circleButton.setActionCommand("circle") ;
      toolbar.add(circleButton) ; 
      /////
      ccircleButton = new JButton("∙") ; 
      ccircleButton.setFont(bfont) ;  
      ccircleButton.createToolTip() ; 
      ccircleButton.setToolTipText("<html><font size=\"+2\">insert ∙ (product) at cursor") ; 
      ccircleButton.setBackground(shadeColor) ; 
      ccircleButton.addActionListener(commandActivator) ; 
      ccircleButton.setActionCommand("ccircle") ;
      toolbar.add(ccircleButton) ; 
      /////
      crossButton = new JButton("⨯") ; 
      crossButton.setFont(bfont) ;  
      crossButton.createToolTip() ; 
      crossButton.setToolTipText("<html><font size=\"+2\">insert ⨯ (cross product) at cursor") ; 
      crossButton.setBackground(shadeColor) ; 
      crossButton.addActionListener(commandActivator) ; 
      crossButton.setActionCommand("cross") ;
      toolbar.add(crossButton) ; 

      toolbar.add(new JLabel("   ")) ;       

      //toolbar.add(new JToolBar.Separator()) ;       
      meetButton = new JButton("∧") ; 
      meetButton.setFont(bfont) ;  
      meetButton.createToolTip() ; 
      meetButton.setToolTipText("<html><font size=\"+2\">insert ∧ (meet) at cursor") ; 
      meetButton.setBackground(shadeColor) ; 
      meetButton.addActionListener(commandActivator) ; 
      meetButton.setActionCommand("meet") ;
      toolbar.add(meetButton) ; 
      /////
      joinButton = new JButton("∨") ; 
      joinButton.setFont(bfont) ;  
      joinButton.createToolTip() ; 
      joinButton.setToolTipText("<html><font size=\"+2\">insert ∨ (join) at cursor") ; 
      joinButton.setBackground(shadeColor) ; 
      joinButton.addActionListener(commandActivator) ; 
      joinButton.setActionCommand("join") ;
      toolbar.add(joinButton) ; 
      /////
      notButton = new JButton("¬") ; 
      notButton.setFont(bfont) ;  
      notButton.createToolTip() ; 
      notButton.setToolTipText("<html><font size=\"+2\">insert ¬ (not) at cursor") ; 
      notButton.setBackground(shadeColor) ; 
      notButton.addActionListener(commandActivator) ; 
      notButton.setActionCommand("not") ;
      toolbar.add(notButton) ; 
      /////
      ifButton = new JButton("→") ; 
      ifButton.setFont(bfont) ;  
      ifButton.createToolTip() ; 
      ifButton.setToolTipText("<html><font size=\"+2\">insert → (imply >) at cursor") ; 
      ifButton.setBackground(shadeColor) ; 
      ifButton.addActionListener(commandActivator) ; 
      ifButton.setActionCommand("if") ;
      toolbar.add(ifButton) ; 
      /////
      fiButton = new JButton("←") ; 
      fiButton.setFont(bfont) ;  
      fiButton.createToolTip() ; 
      fiButton.setToolTipText("<html><font size=\"+2\">insert ← (&lt; < imply) at cursor") ; 
      fiButton.setBackground(shadeColor) ; 
      fiButton.addActionListener(commandActivator) ; 
      fiButton.setActionCommand("fi") ;
      toolbar.add(fiButton) ; 
      /////
      iffButton = new JButton("↔") ; 
      iffButton.setFont(bfont) ;  
      iffButton.createToolTip() ; 
      iffButton.setToolTipText("<html><font size=\"+2\">insert ↔ (equiv) at cursor") ; 
      iffButton.setBackground(shadeColor) ; 
      iffButton.addActionListener(commandActivator) ; 
      iffButton.setActionCommand("iff") ;
      toolbar.add(iffButton) ; 

      //toolbar.add(new JLabel("   ")) ;       

      allButton = new JButton("∀") ; 
      allButton.setFont(bfont) ;  
      allButton.createToolTip() ; 
      allButton.setToolTipText("<html><font size=\"+2\">insert ∀ (for all) at cursor") ; 
      allButton.setBackground(shadeColor) ; 
      allButton.addActionListener(commandActivator) ; 
      allButton.setActionCommand("all") ;
      toolbar.add(allButton) ; 
      /////
      someButton = new JButton("∃") ; 
      someButton.setFont(bfont) ;  
      someButton.createToolTip() ; 
      someButton.setToolTipText("<html><font size=\"+2\">insert ∃ (there exists) at cursor") ; 
      someButton.setBackground(shadeColor) ; 
      someButton.addActionListener(commandActivator) ; 
      someButton.setActionCommand("some") ;
      toolbar.add(someButton) ; 

      //toolbar.add(new JLabel("   ")) ;       

      topButton = new JButton("⊤") ; 
      topButton.setFont(bfont) ;  
      topButton.createToolTip() ; 
      topButton.setToolTipText("<html><font size=\"+2\">insert ⊤ (top) at cursor") ; 
      topButton.setBackground(shadeColor) ; 
      topButton.addActionListener(commandActivator) ; 
      topButton.setActionCommand("top") ;
      toolbar.add(topButton) ; 

      bottomButton = new JButton("⊥") ; 
      bottomButton.setFont(bfont) ;  
      bottomButton.createToolTip() ; 
      bottomButton.setToolTipText("<html><font size=\"+2\">insert ⊥ (bottom) at cursor") ; 
      bottomButton.setBackground(shadeColor) ; 
      bottomButton.addActionListener(commandActivator) ; 
      bottomButton.setActionCommand("bottom") ;
      toolbar.add(bottomButton) ; 

      toolbar.add(new JLabel("   ")) ;       

      intersectButton = new JButton("∩") ; 
      intersectButton.setFont(bfont) ;  
      intersectButton.createToolTip() ; 
      intersectButton.setToolTipText("<html><font size=\"+2\">insert ∩ (intersect) at cursor") ; 
      intersectButton.setBackground(shadeColor) ; 
      intersectButton.addActionListener(commandActivator) ; 
      intersectButton.setActionCommand("intersect") ;
      toolbar.add(intersectButton) ; 

      unionButton = new JButton("∪") ; 
      unionButton.setFont(bfont) ;  
      unionButton.createToolTip() ; 
      unionButton.setToolTipText("<html><font size=\"+2\">insert ∪ (union) at cursor") ; 
      unionButton.setBackground(shadeColor) ; 
      unionButton.addActionListener(commandActivator) ; 
      unionButton.setActionCommand("union") ;
      toolbar.add(unionButton) ; 

      memberButton = new JButton("∈") ; 
      memberButton.setFont(bfont) ;  
      memberButton.createToolTip() ; 
      memberButton.setToolTipText("<html><font size=\"+2\">insert ∈ (member) at cursor") ; 
      memberButton.setBackground(shadeColor) ; 
      memberButton.addActionListener(commandActivator) ; 
      memberButton.setActionCommand("member") ;
      toolbar.add(memberButton) ; 

      emptyButton = new JButton("∅") ; 
      emptyButton.setFont(bfont) ;  
      emptyButton.createToolTip() ; 
      emptyButton.setToolTipText("<html><font size=\"+2\">insert ∅ (empty) at cursor") ; 
      emptyButton.setBackground(shadeColor) ; 
      emptyButton.addActionListener(commandActivator) ; 
      emptyButton.setActionCommand("empty") ;
      toolbar.add(emptyButton) ; 

      toolbar.add(new JLabel("   ")) ;       

      lambdaButton = new JButton("λ") ; 
      lambdaButton.setFont(bfont) ;  
      lambdaButton.createToolTip() ; 
      lambdaButton.setToolTipText("<html><font size=\"+2\">insert λ (lambda) at cursor") ; 
      lambdaButton.setBackground(shadeColor) ; 
      lambdaButton.addActionListener(commandActivator) ; 
      lambdaButton.setActionCommand("lambda") ;
      toolbar.add(lambdaButton) ; 

      piButton = new JButton("Π") ; 
      piButton.setFont(bfont) ;  
      piButton.createToolTip() ; 
      piButton.setToolTipText("<html><font size=\"+2\">insert Π (Pi) at cursor") ; 
      piButton.setBackground(shadeColor) ; 
      piButton.addActionListener(commandActivator) ; 
      piButton.setActionCommand("pi") ;
      toolbar.add(piButton) ; 

      sigmaButton = new JButton("Σ") ; 
      sigmaButton.setFont(bfont) ;  
      sigmaButton.createToolTip() ; 
      sigmaButton.setToolTipText("<html><font size=\"+2\">insert Σ (Sigma) at cursor") ; 
      sigmaButton.setBackground(shadeColor) ; 
      sigmaButton.addActionListener(commandActivator) ; 
      sigmaButton.setActionCommand("sigma") ;
      toolbar.add(sigmaButton) ; 

      toolbar.add(new JLabel("   ")) ;       

      possibleButton = new JButton("◇") ; 
      possibleButton.setFont(bfont) ;  
      possibleButton.createToolTip() ; 
      possibleButton.setToolTipText("<html><font size=\"+2\">insert ◇ (possibly) at cursor") ; 
      possibleButton.setBackground(shadeColor) ; 
      possibleButton.addActionListener(commandActivator) ; 
      possibleButton.setActionCommand("possible") ;
      toolbar.add(possibleButton) ; 

      necessaryButton = new JButton("□") ; 
      necessaryButton.setFont(bfont) ;  
      necessaryButton.createToolTip() ; 
      necessaryButton.setToolTipText("<html><font size=\"+2\">insert □ (necessary) at cursor") ; 
      necessaryButton.setBackground(shadeColor) ; 
      necessaryButton.addActionListener(commandActivator) ; 
      necessaryButton.setActionCommand("necessary") ;
      toolbar.add(necessaryButton) ; 

/////////////////////////////////////////////////////////////////

      togetherPanel.setLayout(new BorderLayout()) ; 
      if (linenumbersFlag) togetherPanel.add("West", linenumbers) ; 
      togetherPanel.add("Center", ta) ; 
      Container c = this.getContentPane() ; 
      c.setLayout(new BorderLayout()) ; 
      c.add("North", toolbar) ; 
      JScrollPane scroller = new JScrollPane(togetherPanel) ; 
      this.vert = scroller.getVerticalScrollBar() ; 
      vert.setUnitIncrement(20) ; 
      vert.setBlockIncrement(40) ; 
      JScrollBar hori = scroller.getHorizontalScrollBar() ; 
      hori.setUnitIncrement(20) ; 
      hori.setBlockIncrement(40) ; 
      c.add("Center", scroller) ; 
      c.add("South", news) ; 

      try { 
         // use settings to set colors ...  
         String clrs = settings.getProperty("backgroundcolor").trim() ; 
         String[] colors = clrs.split(",") ; 
         backgroundColor = new Color(Integer.parseInt(colors[0]),Integer.parseInt(colors[1]),Integer.parseInt(colors[2])) ;
         ta.setBackground(backgroundColor) ;
         // again for selection color
         clrs = settings.getProperty("selectioncolor").trim() ; 
         colors = clrs.split(",") ; 
         selectionColor = new Color(Integer.parseInt(colors[0]),Integer.parseInt(colors[1]),Integer.parseInt(colors[2])) ;
         ta.setSelectionColor(selectionColor) ;
         // decoration preferences 
         clrs = settings.getProperty("commentcolor").trim() ; 
         colors = clrs.split(",") ; 
         commentColor = new Color(Integer.parseInt(colors[0]),Integer.parseInt(colors[1]),Integer.parseInt(colors[2])) ;
         StyleConstants.setForeground(commentForeground,commentColor) ; 
         clrs = settings.getProperty("reservedwordcolor").trim() ; 
         colors = clrs.split(",") ; 
         reservedwordColor = new Color(Integer.parseInt(colors[0]),Integer.parseInt(colors[1]),Integer.parseInt(colors[2])) ;
         StyleConstants.setForeground(reservedWordForeground,reservedwordColor) ; 

      } 
      catch(Exception x)  {    // use default 
         System.out.println("Color not set.") ; 
         ta.setBackground(backgroundColor) ; 
         ta.setSelectionColor(selectionColor) ; 
         // other colors all set when need
      }

      // needs selectionColor 1st
      colorWidget = new FullColorWidget() ; 

      try { 
         // use settings to set some font size 
         fontSize = Integer.parseInt(settings.getProperty("fontsize").trim()) ; 
         Font f = new Font(fontName,fontStyle,fontSize) ; 
         ta.setFont(f) ; 
         linenumbers.setFont(f) ; 
         // still need set menu items 
         switch(fontSize) { 
            case 10 : ten.setSelected(true) ; break ; 
            case 12 : twelve.setSelected(true) ; break ; 
            case 14 : fourteen.setSelected(true) ; break ; 
            case 16 : sixteen.setSelected(true) ; break ; 
            case 18 : eightteen.setSelected(true) ; break ; 
            case 20 : twenty.setSelected(true) ; break ; 
            case 22 : twentytwo.setSelected(true) ; break ; 
            case 24 : twentyfour.setSelected(true) ; break ; 
         }
      }
      catch(Exception x) { 
         fontSize = 14 ; 
         Font f = new Font(fontName,fontStyle,fontSize) ; 
         ta.setFont(f) ; 
         linenumbers.setFont(f) ;         
         fourteen.setSelected(true) ; 
      }

      this.changed = false ;

      /*
         Iconify tools along with the main window.
      */
      this.console_visible = false ; 
      this.colorWidget_visible = false ; 

      this.addWindowListener(new WindowAdapter() { 
         public void windowClosing(WindowEvent w) { 
            exitExam() ; 
         }
         // iconify tools when this is iconified
         public void windowIconified(WindowEvent w) { 
            if (console.isVisible()) {
               console_visible = true ; 
               console.setVisible(false) ; 
            } else console_visible = false ; 
            if (colorWidget.isVisible()) {
               colorWidget_visible = true ; 
               colorWidget.setVisible(false) ; 
            } else colorWidget_visible = false ; 
         }
         // deiconify tools when this is iconified
         public void windowDeiconified(WindowEvent w) { 
            if (console_visible) console.setVisible(true) ; 
            if (colorWidget_visible) colorWidget.setVisible(true) ; 
         }
      }) ; 

      ta.setCaretPosition(0) ;  
      ta.setCaretColor(Color.blue) ;  

      console = new JavaConsole(this) ; 
      console.setSize(500,300) ; 
      console.setLocation(0,0) ; 

      number_of_lines = 1 ;
      linenumbers.append("1 \n") ; 
      linenumbers.append("2 \n") ; 

      pageformat = new PageFormat() ;  // default page format 

   } // end AutoLogEditor(String) constructor



//---------------------------------------------------------------------
//-------- DECORATION ALGORITHMS --------------------------------------
//---------------------------------------------------------------------

   /**
     *  Decorate Java source code. 
     *  Comments are commentColor (default dark greenish).
     *  Java keywords are keywordColor (default dark redish). 
     *  Values are reservedwordColor (default dark bluish).
     *  @value v decorate if true, remove all decoration if false
     */
   void decorateText(boolean v) { 
      int len = ta.getText().length() ;  
      if (len == 0) return ; // no need to decorate an empty string
      if (v) {
         try { 
            decorateReservedWords(0,len-1) ;
            decorateComments(0,len-1) ; 
         } catch(Exception xx) { System.out.println(xx) ; } 
      }
      else {  
         StyledDocument styledDoc = ta.getStyledDocument();
         styledDoc.setCharacterAttributes(0,len-1,fontForeground,true) ; 

      }
   }

   /**
     *  Decorate all slash-star to star_slash comments,  
     *  and slash_slash to line_end comments.
     */
   void decorateComments(int from, int to) { 
      StyledDocument styledDoc = ta.getStyledDocument();
      String source = ta.getText() ; 
      int b = source.indexOf("/*",from) ;  // begin comment
      while(b >= 0 && b <= to) {        
         if (b==0 || (b > 0 && source.charAt(b-1) != '"'))  {    // ignore quoted
            int e = source.indexOf("*/",b) ;  // end of comment 
            if(e >= b) {
               styledDoc.setCharacterAttributes(b,e-b+2,commentForeground,true) ; 
               b = source.indexOf("/*",e+2) ;
            } 
            else b = source.indexOf("/*",b+2) ; // go past unclosed comment
         }
         else b = source.indexOf("/*",b+3) ; // move past this pne
      }
      // Now decorate this kind of comment wherever found
      b = source.indexOf("//",0) ;  // begin this kind of comment
      while(b >= 0 && b <= to) {        
         int e = source.indexOf(10,b) ;  // end of comment line 
         if (b > e) e = source.length()-1 ;  // end of doc w/o linefeed
         styledDoc.setCharacterAttributes(b,e-b + 1,commentForeground,true) ; 
         b = source.indexOf("//",e) ;   // next comment of this type
      }
      b = source.indexOf("%%",0) ;  // begin this kind of comment (Prolog)
      while(b >= 0 && b <= to) {        
         int e = source.indexOf(10,b) ;  // end of comment line 
         if (b > e) e = source.length()-1 ;  // end of doc w/o linefeed
         styledDoc.setCharacterAttributes(b,e-b + 1,commentForeground,true) ; 
         b = source.indexOf("%%",e) ;   // next comment of this type
      }

   }

   /**
     *  Decorate Java reserved words.
     */
   void decorateReservedWords(int from, int to) { 
      StyledDocument styledDoc = ta.getStyledDocument(); 
      String source = ta.getText() ; 
      String resword ; 
      for (int k = 0 ; k < reservedwords.length ; k++) {
         resword = reservedwords[k] ; 
         int len = resword.length() ;
         int p = source.indexOf(resword,from) ; 
         while (p >= 0 && p <= to) { 
               styledDoc.setCharacterAttributes(p,len,reservedWordForeground,true) ; 
            p = source.indexOf(resword,p+len+1) ; 
         }
      }
   } 

   /**
     *  Find the beginning of the line in String holding cursor position p.
     */
   int lineStart(String s, int p) { 
      if (p < 0 || p >= s.length()) return -1 ;  // p out of range
      if (p == 0) return 0 ; // regardless of what char at c 
      int c = p ;       
      while (c > 0) { 
         if (s.charAt(c-1) == 10) return c ; 
         c-- ; 
      }
      return c; 
   }

   /**
     *  Find the end of the line in String holding cursor at p. 
     */
   int lineEnd(String s, int p) { 
      if (p < 0 || p > s.length()) return -1 ;  // p out of range
      int c = p ; 
      while (c < s.length() && s.charAt(c) != 10) c++ ; 
      return c; 
   }

   /**
     *  Find the start of any comment whose scope includes p. 
     */
   Point commentBoundary(String s, int p) { 
      int b = s.indexOf("/*",0) ;   // begin a comment
      while(b >= 0 && b <= p) {     // not after p    
         if (b==0 || (b > 0 && s.charAt(b-1) != '"'))  {    // ignore quoted
            int e = s.indexOf("*/",b) ;  // end of comment 
            if(e >= p) return new Point(b,e) ;    // p falls within this scope 
            else b = s.indexOf("/*",b+2) ;   // go past  an unclosed comment
         }
         else b = s.indexOf("/*",b+3) ; // move past this one
      }
      // p does not lie in the scope of a comment
      return new Point(p,p) ; 
   }


   /**
     *  Rececorate Java keywords or reserved words after insert or remove of text.  
     *  Called by MouseMotion adapter for the text pane.
     *  The DocumentActivator records where inserts and removes have been made.
     *  Only called when source is already decorated,
     *  in an attempt to keep decoration accurate.  THIS DOES NOT WORK YET!
     */
   void redecorate(int a, int b) { 
      String source = ta.getText() ; 
      int doclength = source.length() ; 
      int from = (a <= 0 ? 0 : a)  ,             
          to = (b < doclength ? b : doclength-1) ;  
      //System.out.print (from+":"+to+" , ") ; 
      //System.out.print(lineStart(source,from)+":"+lineEnd(source,to)) ; 
      Point A = commentBoundary(source,lineStart(source,from)), 
            B = commentBoundary(source,lineEnd(source,to)) ; 
      //System.out.println(A.x+":"+B.y) ; 
      StyledDocument styledDoc = ta.getStyledDocument();
      styledDoc.setCharacterAttributes(A.x,B.y-A.x +1,fontForeground,true) ; 
      try { 
         decorateReservedWords(A.x, B.y) ;
         decorateComments(A.x, B.y) ;
      } catch(Exception xx) { System.out.println(xx) ; } 
   } 


/**
  *  Listen for edits that can be undone.
  *  Deliberately ignore style changes because they 
  *  involve decorated source code and are not intended to 
  *  be remembered on the undo stack.  
  */
protected class QUndoableEditListener
                     implements UndoableEditListener {
   public void undoableEditHappened(UndoableEditEvent e) {
      //Remember the edit and update the menus.
      // Do NOT remember style changes -- automatic or under user control
      if (!e.getEdit().getPresentationName().trim().equals("style change")) { 
         undoManager.addEdit(e.getEdit()) ;
         //System.out.println(e.getEdit().getPresentationName().trim().equals("style change")) ; // debug
         undoAction.updateUndoState() ;
         redoAction.updateRedoState() ; 
      }
   }
}

/**
  * Undo Action.
  */
class UndoAction extends AbstractAction {
   public UndoAction() {
       super("Undo");
       setEnabled(false);
   }

   public void actionPerformed(ActionEvent e) {
       try {
           undoManager.undo();
       } 
       catch (CannotUndoException ex) {
          System.out.println("Unable to undo: " + ex);
          ex.printStackTrace();
       }
       updateUndoState();
       redoAction.updateRedoState();
   }

   protected void updateUndoState() {
      if (undoManager.canUndo()) {
           setEnabled(true);
           putValue(Action.NAME, undoManager.getUndoPresentationName());
      } 
      else {
         setEnabled(false);
         putValue(Action.NAME, "Undo");
      }
   }
}


/**
  * Redo Action.
  */
class RedoAction extends AbstractAction {
   public RedoAction() {
      super("Redo");
      setEnabled(false);
   }

   public void actionPerformed(ActionEvent e) {
      try {
         undoManager.redo();
      } 
      catch (CannotRedoException ex) {
         System.out.println("Unable to redo: " + ex);
         ex.printStackTrace();
      }
      updateRedoState();
      undoAction.updateUndoState();
   }

   protected void updateRedoState() {
      if (undoManager.canRedo()) {
         setEnabled(true);
         putValue(Action.NAME, undoManager.getRedoPresentationName());
      } 
      else {
         setEnabled(false);
         putValue(Action.NAME, "Redo");
      }
   }
}

   /** 
   *  Added July 2003.
   *  This can be called by others to load
   *  text from a file into a Java Editor window.
   */ 
   public synchronized void open(File file) { // changed 12/17/14
      try { 
         FileReader fr = new FileReader(file) ;
         BufferedReader br = new BufferedReader(fr) ;
         //BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(file), java.nio.charset.Charset.forName("UTF-16")) ) ;
         StringBuffer sbuffer = new StringBuffer() ; // efficiency!
         String cr = "\n" ; 
         String s = "" ;
         while (s != null) {
            s = br.readLine() ;
            if (s != null) { 
               sbuffer.append(s) ; 
               sbuffer.append(cr) ; 
            }
         }
         br.close() ; 
         s = sbuffer.toString() ; 
         ta.setText(s) ; 
         //previousVersion = s ; 
         renumberLines() ; 
         fileSpec = file ; 
         changed = false ; 
         reportNews() ; 
         ta.setCaretPosition(0) ; 
      }
      catch(Exception e) { 
         //System.out.println("AutoLogEditor (no input file).") ; 
         fileSpec = new File(System.getProperty("user.dir")) ; 
         reportNews() ; 
      }
   }

   /**
     *  Ask user to confirm an unsaved document. 
     *  System.exit(0) is called if no other users
     *  of the VM, else dispose.
     */
   public void exitExam() { 
      if (autologGUI != null) autologGUI.doExit() ;
      if (!changed || ta.getText().equals("") )    { 
         System.exit(0) ; 
      }
      else { 
         Toolkit.getDefaultToolkit().beep() ; 
         int response = 
            JOptionPane.showConfirmDialog(this,
                "Save before exit?",
                "Document has changed" ,
                JOptionPane.YES_NO_CANCEL_OPTION,
                JOptionPane.WARNING_MESSAGE,
                null) ; 
         if(response==0) { 
            fileActivator.save() ; 
            System.exit(0) ; 
         }
         else if (response==1) { 
            System.exit(0) ; 
         }  
      } 
   }

   /**
   *  Added method July 2003.
   *  Set the text.  Intended to be used by other 
   *  application; first create new editor, then set
   *  the text from the other application.
   */
   public void setText(String text) { 
      ta.setText(text) ; 
      reportNews() ; 
   } 

   /**
   *  Maintain news in label at bottom bar of editor. 
   *  Notify user if text has CHANGED.  Tell user how many lines.
   */
   private void reportNews() { 
      this.news.setText((changed ? String.valueOf(number_of_lines)+" lines "+ " ... and CHANGED" : String.valueOf(number_of_lines)+" lines ")) ; 
      if (fileSpec.isDirectory()) 
         this.setTitle(fileSpec.toString() + File.separator +" [unsaved]") ; 
      else 
         this.setTitle(fileSpec.toString()) ; 
   } 

   /**
   *  Increase number of lines for more text.
   *  Uses append for text in linenumbers.
   *  Called via DocumentActivator. 
   *  Smoother than renumberLines().
   */
   private void updateLineNumbering() { 
      int n = numberOfLines(ta.getText()) ; 
      // if more lines are needed
      if (n > number_of_lines) { 
         // Use append so as not to upset focus
         for(int j = number_of_lines+2; j <= n+1; j++) 
            linenumbers.append(String.valueOf(j) + " \n") ;
         number_of_lines = n ; 
      }
   }

   /**
   *  Renumber lines absolutely.
   *  Called by MouseAdapter for linenumbers  
   *  and FileActivator (New and Open, etc.).
   */
   private void renumberLines() { 
      int n = numberOfLines(ta.getText()) ; 
      if (n != number_of_lines) { 
         String numbers = "" ; 
         for(int k = 1 ; k <= n+1 ; k++) 
            numbers += String.valueOf(k) + " \n" ;
         linenumbers.setText(numbers) ; 
         number_of_lines = n ;  
      }
   }

   /**
   *  See how many lines text has.
   *  Count 10's.  This seems to work best, 
   *  all things considered.
   */
   private int numberOfLines(String text) {
      if (text.equals("")) return 0 ; 
      int count = 0 ; 
      for(int i = 0 ; i < text.length() ; i++) 
         if ((int)text.charAt(i) == 10) count++ ; 
      // text ends with 10
      if (text.charAt(text.length()-1) == 10) return count ; 
      // text does not end with 10
      else return count+1 ;
   }

   /**
   *  Compute beginning caret position corresponding to numlines.
   */
   private int findCaretPosition(int numlines) { 
      if (numlines == 1) return 0 ; // start of first line
      String text = ta.getText() ; 
      int p = 0 ;          // text position
      int count10s = 0 ;   // how many returns
      while (p < text.length()) {
         if ((int)text.charAt(p) == 10) 
            count10s++ ;  
         if (count10s == numlines - 1) return p+1 ;
 
        p++ ; 
      } 
      return 0 ; 
   }

   /**
   *  Highlight a line of text.
   */
   private void hiliteLine(int line) { 
      int start = findCaretPosition(line) ; 
      int end = start ; 
      String text = ta.getText() ; 
      while ((end < text.length()-start) && (text.charAt(end) != 10) ) end++ ;
      ta.select(start, end) ; 
   }

   /**
   *  Inner class fulfills usual console duties. 
   *  Used by CommandActivator for compiling and running.
   */
   class JavaConsole extends JFrame 
                      implements ActionListener { 
      JTextArea output ;  
      JButton gt, clr ; 
      JTextField input ; 
      JPopupMenu consolepopedit ; 

      PrintWriter writer ; 
      AutoLogEditor jed ; 

      JavaConsole(AutoLogEditor jed) { 
         super("AutoLog Console") ; 
         this.jed = jed ; 
         output = new JTextArea() ;  
         output.setFont(new Font("monospaced", Font.PLAIN, activatorFor.fontSize)) ;  
         output.setBackground(backgroundColor) ; 
         output.setSelectionColor(selectionColor) ; 
         Container c = this.getContentPane() ; 
         c.setLayout(new BorderLayout()) ;
         c.add("Center", new JScrollPane(output)) ; 
         JPanel p = new JPanel() ; 
         p.setLayout(new BorderLayout()) ; 

         gt = new JButton("Goto") ; 
         gt.createToolTip() ; 
         gt.setToolTipText("Select error line number and then press this button...") ;
         gt.setActionCommand("goto") ; 
         gt.addActionListener(this) ; 

         input = new JTextField() ; 
         input.setActionCommand("input") ; 
         input.addActionListener(this) ;

         JPanel p1 = new JPanel() ; 
         p1.setLayout(new BorderLayout()) ; 
         JLabel inLabel = new JLabel("  Inputs: ") ; 
         inLabel.createToolTip() ; 
         inLabel.setToolTipText("Supply inputs here for running program ...") ;
         p1.add("West", inLabel) ; 
         p1.add("Center",input) ; 

         p.add("West", gt) ; 
         p.add("Center", p1) ; 
         //p.add("East", clr) ; 
         c.add("South", p) ; 
         this.addWindowListener(new WindowAdapter() { 
            public void windowClosing(WindowEvent e) { 
               setVisible(false) ; 
               console_visible = false ; 
            } 
         }) ;

         // pop this up with right click
         consolepopedit = new JPopupMenu() ; 
         JMenuItem consolecopy = new JMenuItem("Copy") ; 
         consolecopy.addActionListener(this) ; 
         consolecopy.setActionCommand("consolecopy") ; 
         consolepopedit.add(consolecopy) ;  

         output.addMouseListener(new MouseAdapter() { 
            public void mouseReleased(MouseEvent e) { 
               if (e.isPopupTrigger())  // popup now
                  consolepopedit.show(output,e.getX(),e.getY()) ; 
            } 
            public void mousePressed(MouseEvent e) { 
               if (e.isPopupTrigger())  // popup now
                  consolepopedit.show(output,e.getX(),e.getY()) ; 
            } 
         }) ;

      }

      public void actionPerformed(ActionEvent e) { 
         if (e.getActionCommand().equals("input") && (writer != null)) { 
            updateConsoleWindow("\n") ; 
            writer.println(input.getText()) ;
            writer.flush() ; 
         }            
         else if (e.getActionCommand().equals("consolecopy")) { 
            output.copy() ; 
         }
         else if (e.getActionCommand().equals("clear")) { 
             output.setText("") ; 
             input.setText("") ; 
         } 
         else if (e.getActionCommand().equals("goto")) { 
             try { 
                jed.hiliteLine(Integer.parseInt(output.getSelectedText())) ; 
             } catch(Exception x) {}
         }
      }
      
      void updateConsoleWindow(String s) { 
         output.append(s) ; 
      }
   } // end JavaConsole inner class

   // INNER JeditorPrinter class
   /**
   *  This inner class describes how to "print" a 
   *  text document.  Printing can be a screen preview or on paper.
   *  WYSIWYG --   Including line numbers if activated.
   *  "Jed" used for several Java editors since 1990's.
   */
   class JeditorPrinter implements Printable {
      Font font ; 

      JeditorPrinter() { 
         this.font = ta.getFont() ; 
      }
   
      /**
      *  Calculate how many lines will fit on a page.
      */
      int numberOfLinesPerPage(Graphics g, PageFormat format) {   
         FontMetrics fontmetrics = g.getFontMetrics(font) ; 
         return 
            (int)(format.getImageableHeight() / fontmetrics.getHeight()) ; 
      }

      /**
      *  Calculate how many pages there are.
      */
      int totalNumberOfPages(Graphics g, PageFormat format) { 
         StringTokenizer tok = new StringTokenizer(ta.getText(),"\n") ; 
         int numlines = numberOfLines(ta.getText()) ;
         return (int)Math.ceil((double)numlines/(double)numberOfLinesPerPage(g,format)) ; 
      } 

      /**
      *  Calculate the starting line number of a page.
      */
      int startingLine(Graphics g, PageFormat format, int page) { 
         return page * numberOfLinesPerPage(g,format) ; 
      }

      /**
      *  Calculate how many characters will fit per line.
      */
      int numberCharsPerLine(Graphics g, PageFormat format) { 
         FontMetrics metrics = g.getFontMetrics(font) ; 
         return 
            (int)(format.getImageableWidth() / metrics.charWidth('A')) ;
      }

      /** 
      *  Return a string representation of an int
      *  left-justified over a width of w.
      *  Use this to number printed lines with 
      *  uniform column spacing.
      */ 
      String format(int n, int w) {
         String s = String.valueOf(n) ;
         while (s.length() < w) s = s + " " ;
         return s ;
      }

      /**
      *  For preview or paper ...
      */
      public int print(Graphics g,
                       PageFormat format,
                       int pagenumber)   {
         int totalnumberofpages = totalNumberOfPages(g,format) ; 
         if (pagenumber > totalnumberofpages) return Printable.NO_SUCH_PAGE ; 
         g.setFont(font) ; 
         FontMetrics metrics = g.getFontMetrics(font) ; 
         int fontheight = metrics.getHeight() ; 
         int leftmargin = (int)format.getImageableX() ; 
         int topmargin = (int)format.getImageableY() ; 
         int numbercharsperline = numberCharsPerLine(g,format)- (int)(0.5)*leftmargin ; 
         int startingline = startingLine(g,format,pagenumber) ;
         int numberoflinesperpage = numberOfLinesPerPage(g,format) ;
         String source = ta.getText() ; 
         if (source.equals("")) return Printable.NO_SUCH_PAGE ; 
         int counter = 0 ;     // count lines
         int index = 0 ;       // index of character in source
         int len = source.length() ; 
         // thumb through beginning lines, up to starting line ...
         while (counter < startingline) { 
            if (source.charAt(index) == 10) counter++ ;
            index++ ; 
         } 
         // now draw this page starting with startingline 
         String line ; 
         int totalnumberoflines = numberOfLines(source) ; 
         int cwidth = String.valueOf(totalnumberoflines).length() ; 
         int beginning = index ;     // beginning of line 
         while((index < len) && (counter < startingline + numberoflinesperpage)) { 
            beginning = index ; 
            // find end of line
            while ((index < len) && (source.charAt(index) != 10)) index++ ; 
            // for the last line
            //if (index == len) index-- ; 
            counter++ ; 
            if (beginning == index) line = " " ; 
            else line = source.substring(beginning,index) ; 
            int last = Math.min(line.length(),numbercharsperline) ; 
            String prefix = (linenumbersFlag ? format(counter,cwidth) : "") ;  
            g.drawString(prefix + 
                                 " " + line.substring(0,last), 
                            leftmargin,
                            topmargin + (counter-startingline)*fontheight  
                         ) ; 
            index++ ; 
         }
         return Printable.PAGE_EXISTS ;
      }
   }

//---------------------------------------------------------------------
//-------- MAIN -------------------------------------------------------
//---------------------------------------------------------------------

   
   /** 
     *  ... from the command line ... generates Exception at line#1490
     */ 
   public static void main(String[] args) {       
      //System.out.println(UIManager.getCrossPlatformLookAndFeelClassName()) ; 
      try { 
         Properties settings = new Properties() ; 
         settings.load(
           new FileInputStream(
            System.getProperty("user.home")+File.separator+"AutoLogEditor.settings" )  ) ; 
         UIManager.setLookAndFeel(settings.getProperty("lookandfeel")); // Exception ?
      }
      catch (Exception x) { 
         try { 
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName()); 
         }
         catch (Exception xx) {        
         } 
      } 
      final AutoLogEditor tje =  new AutoLogEditor("AutoLog Script Editor "+version) ; 

      try { 
         if (args.length > 0) 
         tje.open(new File(args[0])) ; // 8/19/13
      }
      catch(Exception x) {
         System.out.println(x) ;
         System.out.println("AutoLogEditor -- no input file") ;
      }

      //tje.setVisible(true) ;  
      // Wait for layout.  1/3/04
      Thread shows = new Thread() { 
         public void run() { 
            tje.setVisible(true) ; 
         }
      } ;
      EventQueue.invokeLater(shows) ;  

/*
      // Explore available system fonts ...
      GraphicsEnvironment ge = 
          GraphicsEnvironment.getLocalGraphicsEnvironment() ; 
      Font[] fonts = ge.getAllFonts() ; 
      for (int i = 0 ; i < fonts.length ; i++) { 
         Font f = (Font)fonts[i] ; 
         System.out.println(f.getFontName()) ; 
      }
*/
      System.out.println(System.getProperty("os.name")) ; 
      System.out.println(UIManager.getSystemLookAndFeelClassName()) ; 

   } //UIManager.getSystemLookAndFeelClassName()

//---------------------------------------------------------------------
//-------- COLOR CHOOSER ----------------------------------------------
//---------------------------------------------------------------------

/** 
  *  Use a JColorChooser to allow editor to 
  *  assign any color to its attribute.  
  */
class FullColorWidget extends JFrame
                              implements ChangeListener,
                                         ActionListener {

   private JRadioButton backgroundButton, selectionButton, fontButtom, 
                        commentButton, keywordButton, reservedwordButton ;
   JColorChooser jcc; 

   public FullColorWidget() { 
      super("AutoLogEditor Colors") ; 
      jcc = new JColorChooser(Color.white);
      jcc.getSelectionModel().addChangeListener(this);
      jcc.setBorder(BorderFactory.createTitledBorder("Select Color"));
      this.setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);

      JPanel modePanel = new JPanel() ; 
      modePanel.setBorder(BorderFactory.createTitledBorder("Select Mode"));
      modePanel.setLayout(new GridLayout(2,2)) ; 
      backgroundButton = new JRadioButton("Background") ; 
      backgroundButton.addActionListener(this) ; 
      selectionButton = new JRadioButton("Selection") ; 
      selectionButton.addActionListener(this) ; 
      commentButton = new JRadioButton("Comments") ; 
      commentButton.addActionListener(this) ; 
//      keywordButton = new JRadioButton("Keywords") ; 
//      keywordButton.addActionListener(this) ; 
      reservedwordButton = new JRadioButton("Reserved Words") ; 
      reservedwordButton.addActionListener(this) ; 

      ButtonGroup bg = new ButtonGroup() ; 
      bg.add(backgroundButton) ; 
      bg.add(selectionButton) ; 
      bg.add(commentButton) ; 
      bg.add(keywordButton) ; 
      bg.add(reservedwordButton) ; 
      modePanel.add(backgroundButton) ; 
      modePanel.add(selectionButton) ; 
      modePanel.add(commentButton) ; 
//      modePanel.add(keywordButton) ; 
      modePanel.add(reservedwordButton) ; 

      Container c = this.getContentPane() ;
      c.setLayout(new BorderLayout()) ; 
      c.add("North", modePanel) ; 
      c.add("Center",jcc); 
      this.pack();
   }

   public void stateChanged(ChangeEvent e) {
      Color newColor = jcc.getColor(); 
      if (backgroundButton.isSelected()) { 
         backgroundColor = newColor ; 
         ta.setBackground(newColor) ; 
         console.output.setBackground(newColor) ; 
      }
      else if (selectionButton.isSelected()) {
         selectionColor = newColor ; 
         ta.setSelectionColor(newColor) ; 
      }
      else if (commentButton.isSelected()) {
         commentColor = newColor ; 
         StyleConstants.setForeground(commentForeground,commentColor) ; 
         if (decorateTextFlag) { 
            decorateText(true) ; 
         }
      }
//      else if (keywordButton.isSelected()) {
//         keywordColor = newColor ; 
//         StyleConstants.setForeground(keyWordForeground,keywordColor) ; 
//         if (decorateTextFlag) { 
//            decorateText(true) ; 
//         }
//      }
      else if (reservedwordButton.isSelected()) {
         reservedwordColor = newColor ; 
         StyleConstants.setForeground(reservedWordForeground,reservedwordColor) ; 
         if (decorateTextFlag) { 
            decorateText(true) ; 
         }
      }
      else {}
   }

   public void actionPerformed(ActionEvent e) { 
      if (e.getSource().equals(backgroundButton)) { 
         jcc.setColor(backgroundColor) ; 
      }
      else if (e.getSource().equals(selectionButton)) { 
         jcc.setColor(selectionColor) ; 
      }
      else if (e.getSource().equals(commentButton)) { 
         jcc.setColor(commentColor) ; 
      }
//      else if (e.getSource().equals(keywordButton)) { 
//         jcc.setColor(keywordColor) ; 
//      }
      else if (e.getSource().equals(reservedwordButton)) { 
         jcc.setColor(reservedwordColor) ; 
      }
      else {}
   }

}


//=========================
// INNER ACTIVATOR CLASSES
//=========================

/**
  * Inner Activator watches for text changes.
  */
class DocumentActivator implements DocumentListener { 
   public void insertUpdate(DocumentEvent e) { 
      changed = true ; 
      updateLineNumbering() ; 
      // watch for changes for decoration
      if (decorateTextFlag) { 
         int p = e.getOffset() ; 
         int g = e.getLength() ;
         if (redecorate.x < 0 ) {
            redecorate.x = p ; 
            redecorate.y = p+g ; 
         }
         else {
            redecorate.x = Math.min(redecorate.x,p) ; 
            redecorate.y = Math.max(p+g,redecorate.y) ; 
         }
      }
      reportNews() ; 
      //System.out.println("insert "+e.getOffset()+":"+e.getLength()) ; 
   }
   public void changedUpdate(DocumentEvent e) {
      //changed = true ;  // text has not changed
      reportNews() ; 
      updateLineNumbering() ; 
      //System.out.println("changed "+e.getOffset()+":"+e.getLength()) ; 
   }
   public void removeUpdate(DocumentEvent e) {
      changed = true ; 
      renumberLines() ; 
      //updateLineNumbering() ; 
      // watch for changes for decoration
      if (decorateTextFlag) { 
         int p = e.getOffset() ; 
         int g = e.getLength() ;
         if (redecorate.x < 0) {
            redecorate.x = p ; 
            redecorate.y = p+g ; 
         }
         else {
            redecorate.x = Math.min(redecorate.x,p) ; 
            redecorate.y = Math.max(p+g,redecorate.y) ; 
         }
      }
      reportNews() ; 
      //System.out.println("remove "+e.getOffset()+":"+e.getLength()) ; 
   }
}
  // end DocumentActivator

/**
  *  Inner FileActivator serves file menu functions.
  */
class FileActivator implements ActionListener { 
   public void actionPerformed(ActionEvent e) {
      String command = e.getActionCommand() ; 
      if (command.equals("new")) {
         // changed 25 June 2009, "new" now means "new window"
         final AutoLogEditor newed = new AutoLogEditor("AutoLogEditor") ; 
         // wait for layout
         Thread shows = new Thread() { 
            public void run() { 
               newed.setVisible(true) ; 
            }
         } ;
         EventQueue.invokeLater(shows) ;  
      }
      else if (command.equals("save")) save() ;
      else if (command.equals("saveas")) saveAs() ;
      else if (command.equals("open")) 
         if (!changed) open() ; 
         else {  
            Toolkit.getDefaultToolkit().beep() ; 
            int response = 
               JOptionPane.showConfirmDialog(activatorFor,
                   "Save before open?",
                   "Document has changed" ,  
                   JOptionPane.YES_NO_CANCEL_OPTION, 
                   JOptionPane.WARNING_MESSAGE, 
                   null) ; 
            if(response==0) { 
               save() ; 
               open() ; 
            }
            else if (response==1) { 
               open() ;
            } 
           
         }
      //else if (command.equals("exit")) { 
      //   activatorFor.exitExam() ;
      //}
      activatorFor.reportNews() ; 
   }

   // Open and read file of text.   OPEN
   synchronized void open() {
      String cr = "\n" ; 
      if (fileChooser == null) fileChooser = new JFileChooser() ;       
      fileChooser.setSelectedFile(fileSpec) ;   
      int result = fileChooser.showOpenDialog(activatorFor) ; 
      if (result == JFileChooser.APPROVE_OPTION) { 
         File newfileSpec = fileChooser.getSelectedFile() ; 
         try {     
            FileReader fr = new FileReader(newfileSpec) ;
            BufferedReader br = new BufferedReader(fr) ;
            StringBuffer sbuffer = new StringBuffer() ; // efficiency!
            String s = "" ; 
            while (s != null) {
               s = br.readLine() ;  // the next line
               if (s != null) { 
                  sbuffer.append(s) ; 
                  sbuffer.append(cr) ; 
               }
            }
            br.close() ;
            s = sbuffer.toString() ; 
            ta.setText(s) ; 
            //previousVersion = s ; 
            renumberLines() ; 
            changed = false ; 
            fileSpec = newfileSpec ; 
            reportNews() ; 
         }
         catch (Exception e) { 
            Toolkit.getDefaultToolkit().beep() ; 
            JOptionPane.showMessageDialog(activatorFor, 
                                          e.toString(),
                                          "Bad file name ...",
                                          JOptionPane.WARNING_MESSAGE) ;       
         }
      } 
      ta.setCaretPosition(0) ; 
   }

   synchronized void save() { 
      // fileSpec will be current directory if text not yet saved
      if (!fileSpec.isDirectory()) { 
         try { 
            FileWriter fw = new FileWriter(fileSpec) ;
            fw.write(ta.getText()) ;
            fw.close() ; 
         }
         catch (IOException e) { System.out.println(e) ; }
         changed = false ; 
         reportNews() ; 
      } 
      else saveAs() ; 
   }

   synchronized void saveAs() {
      if (fileChooser == null) fileChooser = new JFileChooser() ;       
      fileChooser.setSelectedFile(fileSpec) ;   
      int result = fileChooser.showSaveDialog(activatorFor) ; 
      if (result == JFileChooser.APPROVE_OPTION) { 
         boolean saveit = true ; 
         File newfileSpec = fileChooser.getSelectedFile() ; 
         // ask before saving over existing file ...
         if (newfileSpec.isFile()) { 
            Toolkit.getDefaultToolkit().beep() ; 
            int response = 
               JOptionPane.showConfirmDialog(activatorFor,
                   newfileSpec + " already exists.  Save anyway?",
                   "Existing file ... " ,
                   JOptionPane.YES_NO_OPTION) ; 
             if (response == 1) saveit = false ; 
         }
         if (saveit) { 
            try { 
               FileWriter fw = new FileWriter(newfileSpec) ;
               fw.write(ta.getText()) ;
               fw.close() ; 
               fileSpec = newfileSpec ;  // update only now
               changed = false ; 
               reportNews() ; 
            } 
            catch(Exception x) { 
               Toolkit.getDefaultToolkit().beep() ; 
               JOptionPane.showMessageDialog(activatorFor, 
                                             x.toString(),
                                             "Bad file name ...",
                                             JOptionPane.WARNING_MESSAGE) ;       
            }
         }  
      }
   }
}   
  //end FileActivator

/**
  *  Inner OptionsActivator serves Options menu.
  */
class OptionsActivator implements 
                       ActionListener {

   public void actionPerformed(ActionEvent e) { 
      String command = e.getActionCommand() ; 
      System.out.println(command) ;
      if (command.equals("Toolbar")) {
         if (toolbar.isVisible()) { 
            toolbar.setVisible(false) ; 
         }
         else { 
            toolbar.setVisible(true) ; 
         }
      }   
      else if (command.equals("showlinenumbers")) {  
         if (linenumbersFlag) { 
            togetherPanel.remove(linenumbers) ; 
            linenumbersFlag = false ; 
            togetherPanel.doLayout() ; 
         } 
         else { 
            togetherPanel.add("West", linenumbers) ; 
            linenumbersFlag = true ; 
            togetherPanel.doLayout() ; 
         } 
      } 
      else if (command.equals("decorate")) { 
         if (decorateTextFlag) { 
            decorateText(false) ; 
            decorateTextFlag = false ; 
         } 
         else { 
            decorateText(true) ; 
            decorateTextFlag = true ; 
         }
      }
      else if (command.indexOf("fontsize") == 0) { 
         // parse font size from action command
         fontSize = Integer.parseInt(command.substring(8, command.length())) ; 
         Font f = new Font(fontName, fontStyle, fontSize) ; 
         ta.setFont(f) ;
         // use same font for line numbers to be sure things line up
         linenumbers.setFont(f) ;
      }
      else if (command.equals("colors")) { 
         colorWidget.setVisible(true) ; 
      }

      else if (command.equals("save_settings")) { 
         // Set up for JOptionPane.showOptionDialog 
         Object[] contents = new Object[6] ; 
         String message = new String("Save settings to: " + 
                         System.getProperty("user.home") + 
                           File.separator + "AutoLogEditor.settings") ; 
         String separator = "... save window location & size, colors, commands and  " ; 
         contents[0] = message ; 
         contents[1] = separator ;
         contents[2] = new String("Your Look and Feel preference:  ") ; 
         JRadioButton crossPLF = 
           new JRadioButton(UIManager.getCrossPlatformLookAndFeelClassName() + " (cross-platform LAF) ") ; 
         JRadioButton systemLF = 
           new JRadioButton(UIManager.getSystemLookAndFeelClassName() + " (your system LAF) ") ; 
         ButtonGroup gp = new ButtonGroup() ; 
         gp.add(crossPLF) ; 
         gp.add(systemLF) ; 
         //System.out.println("?"+UIManager.getLookAndFeel().getID()+"?") ;
         if (UIManager.getLookAndFeel().getID().equals("Metal"))
            crossPLF.setSelected(true) ; 
         else systemLF.setSelected(true) ; 
         contents[3] = crossPLF ;
         contents[4] = systemLF ; 
         contents[5] = "A new Look and Feel only appears the next time AutoLogEditor starts." ; 
         Object[] options = { "Ok", "Cancel" } ; 
         int result = 
           JOptionPane.showOptionDialog(activatorFor, 
                                          contents, 
                                          "Save user settings ", 
                                          JOptionPane.DEFAULT_OPTION, 
                                          JOptionPane.QUESTION_MESSAGE,
                                          null, options, options[0]); 
         if (result == 0) { 
            //System.out.println("You chose to save the settings!") ; 
            Properties now = new Properties() ; 
            Dimension d = activatorFor.getSize() ; 
            now.put("width",String.valueOf(d.width)) ; 
            now.put("height",String.valueOf(d.height)) ; 
            Point p = activatorFor.getLocation() ; 
            now.put("x",String.valueOf(p.x)) ; 
            now.put("y",String.valueOf(p.y)) ; 
            now.put("fontsize",
                     String.valueOf(ta.getFont().getSize())) ; 
//            now.put("compilecommand", activatorFor.commandActivator.compilecommand) ; 
//            now.put("runcommand", activatorFor.commandActivator.runcommand) ; 
//            now.put("runPYcommand", activatorFor.commandActivator.runPYcommand) ; 
            now.put("backgroundcolor",  String.valueOf(backgroundColor.getRed()) + "," 
                             + String.valueOf(backgroundColor.getGreen()) + "," 
                               + String.valueOf(backgroundColor.getBlue())  )   ;
            now.put("selectioncolor",  String.valueOf(selectionColor.getRed()) + "," 
                             + String.valueOf(selectionColor.getGreen()) + "," 
                               + String.valueOf(selectionColor.getBlue())  )   ;
            now.put("commentcolor",  String.valueOf(commentColor.getRed()) + "," 
                             + String.valueOf(commentColor.getGreen()) + "," 
                               + String.valueOf(commentColor.getBlue())  )   ;
//            now.put("keywordcolor",  String.valueOf(keywordColor.getRed()) + "," 
//                             + String.valueOf(keywordColor.getGreen()) + "," 
//                               + String.valueOf(keywordColor.getBlue())  )   ;
            now.put("reservedwordcolor",  String.valueOf(reservedwordColor.getRed()) + "," 
                             + String.valueOf(reservedwordColor.getGreen()) + "," 
                               + String.valueOf(reservedwordColor.getBlue())  )   ;

            if (crossPLF.isSelected())
               now.put("lookandfeel",UIManager.getCrossPlatformLookAndFeelClassName() ) ; 
            else 
               now.put("lookandfeel",UIManager.getSystemLookAndFeelClassName() ) ; 
            try { 
               FileOutputStream fos = 
                  new FileOutputStream(
                    System.getProperty("user.home")+File.separator+"AutoLogEditor.settings") ; 
               now.store(fos, "AutoLogEditor user properties") ; 
            }
            catch(Exception x) {  
               System.out.println(x) ; 
            }
         }
      }
      else if (command.equals("info")) { // NOT used by OptionsActivator
         Object[] message = {
            "User can open file by dropping it on menubar.", 
            "   ",
            "AutoLogEditor compiles Java source file to the file's opening directory",
            "unless user modifies \"Other Switches\" field in popup.",
            "   ",
            "fisher.r.john@gmail.com  (mention AutoLogEditor on subject line)",
            "   "
         } ;
         JOptionPane.showMessageDialog(activatorFor,message,"About AutoLogEditor",JOptionPane.INFORMATION_MESSAGE) ;
      }
      else if (command.equals("up")) {
         try {
           //System.out.println(vert.getValue()) ;
           vert.setValue(vert.getValue()-50) ;
         } catch (Exception x) {  System.out.println(x) ; }
      }
      else if (command.equals("tidy")) {
         // Set font 14pt, no linenumbers
         if (linenumbersFlag) { 
            togetherPanel.remove(linenumbers) ; 
            linenumbersFlag = false ; 
         }
         Font f = new Font(fontName, fontStyle, 14) ; 
         fourteen.setSelected(true) ; 
         ta.setFont(f) ;
         togetherPanel.doLayout() ; 
      }
      else if (command.equals("big")) {
         // Set font 24pt, linenumbers
         if (!linenumbersFlag) { 
            togetherPanel.add("West", linenumbers) ;  
            linenumbersFlag = true ; 
         }
         Font f = new Font(fontName, fontStyle, 24) ; 
         twentyfour.setSelected(true) ; 
         ta.setFont(f) ;
         linenumbers.setFont(f) ;
         togetherPanel.doLayout() ; 
      }
      else if (command.equals("boundary")) {
         // right margine marker  THIS MIGHT BE IMPROVED!
         String bndry = 
            "                                                         █\n" ;
         //int cp = ta.getCaretPosition() ;
         ta.setText(bndry+ta.getText()) ;
         ta.setCaretPosition(bndry.length()) ;
      }
   }    
} 
   //end OptionsActivator

/** 
  *  Inner EditActivator activates
  *  COPY, CUT, PASTE, UNDO/REDO and popup edit menu manager.
  *  Note that this Activator listens for two 
  *  different kinds of events.
  *  
  *  Undo reverts to the version of the text since the last
  *  cut or paste, Redo restores what Undo has undone. 
  *
  *  FIND and FINDAGAIN have a simple implementation here.
  */
class EditActivator implements ActionListener, MouseListener { 
   String stringToFind ; 
   JTextField whatToFind  ; 

   EditActivator() { 
      stringToFind = "" ; 
      //previousVersion = "" ;
      whatToFind = new JTextField(stringToFind) ; 
   }
   // actions to edit 
   public void actionPerformed(ActionEvent e) { 
      String command = e.getActionCommand() ; 
      if (command.equals("selectall")) { 
         ta.requestFocus() ; 
         ta.setSelectionStart(0) ; 
         ta.setSelectionEnd(ta.getText().length()) ; 
      }
      else if (command.equals("copy")) { 
         ta.copy() ; 
      }
      else if (command.equals("cut")) { 
         String str = ta.getSelectedText() ;
         if (str != null) {
            //previousVersion = ta.getText().substring(0) ; // a copy
            ta.cut() ; 
            //updateLineNumbering() ;  
            changed = true ; 
         }
      }
      else if (command.equals("paste")) { 
         try { 
            //previousVersion = ta.getText().substring(0) ; // a copy
            ta.paste() ; 
            //updateLineNumbering() ; 
            changed = true ; 
         } catch(Exception exc) {}
      }
      else if (command.equals("find")) { 
         String s = 
            JOptionPane.showInputDialog(activatorFor,"Find (case sensitive): ", stringToFind) ;
         if (s != null) { 
            stringToFind = s ; 
            int start = ta.getText().indexOf(stringToFind,ta.getCaretPosition()) ; 
            if (start != -1)  { // then found 
               ta.select(start,start+stringToFind.length()) ; 
               ta.requestFocus() ; 
            }
            else { 
               JOptionPane.showMessageDialog(activatorFor, 
                   "Finished search -- to end of document ! ",
                   "Find "+ stringToFind,
                   JOptionPane.INFORMATION_MESSAGE) ;
            }
         }
      }
      else if (command.equals("findagain")) { 
         try { 
            int start = ta.getText().indexOf(stringToFind,ta.getCaretPosition()) ; 
            if (start != -1) 
               ta.select(start,start+stringToFind.length()) ; 
            else { 
               JOptionPane.showMessageDialog(activatorFor, 
                      "Finished search -- end of document ! ",
                      "Find "+ stringToFind,
                      JOptionPane.INFORMATION_MESSAGE) ;
            }
         }
         catch(Exception x) {} 
      } 
      else if (command.equals("indent")) { 
         String sindent = "5" ;
         sindent = 
            JOptionPane.showInputDialog(activatorFor,"Indent selected text: ", sindent) ;
         try { 
            if (sindent != null) { 
               int indent = Integer.parseInt(sindent) ;
               // System.out.println(indent) ;  // ok
               int start = ta.getSelectionStart(),
                   end = ta.getSelectionEnd();
               String prefix = "" ;
               for(int i=0; i <indent ; i++) prefix = prefix + " " ;                
               String selection = ta.getSelectedText() ;
               //System.out.println(selection) ;  // ok
               String[] lines = selection.split("\n") ; 
               StringBuilder builder = new StringBuilder() ; 
               for(int j = 0 ; j < lines.length ; j++) {
                  //System.out.println(lines[j]) ;  
                  builder.append(prefix) ; 
                  builder.append(lines[j]) ;
                  builder.append("\n") ;
               }
               //System.out.println(builder.toString()) ;   
               StringBuilder ta_build = new StringBuilder(ta.getText()) ;
               ta_build.replace(start,end,builder.toString()) ;  
               ta.setText(ta_build.toString()) ; 
               // activatorFor.hiliteLine(end) ;   // ??? jumps away from indent
               JOptionPane.showMessageDialog(activatorFor, 
                      "Undo twice to revert",
                      " Indented ",
                      JOptionPane.INFORMATION_MESSAGE) ;
     
            }
         }
         catch(Exception x) {} 
      

      }
      else if (command.equals("goline")) {
         String sgoline = "" ;
         sgoline = 
            JOptionPane.showInputDialog(activatorFor,"Goto line: ", sgoline) ;
         try {  // goes to line but only highlights going UP ?
            if (sgoline != null) { 
               //activatorFor.hiliteLine(activatorFor.number_of_lines - 1) ; 
               activatorFor.hiliteLine(Integer.parseInt(sgoline)) ; 
            }
          }
         catch(Exception x) {} 
      } 
   } 

   // pop up edit menu
   public void mouseReleased(MouseEvent me) { 
      if (me.isPopupTrigger()) // popup now
        popedit.show(ta,me.getX(),me.getY()) ; 
   }
   public void mousePressed(MouseEvent me) { 
      if (me.isPopupTrigger()) // popup now
        popedit.show(ta,me.getX(),me.getY()) ; 
   }

   public void mouseClicked(MouseEvent me) {}
   public void mouseEntered(MouseEvent me) {}
   public void mouseExited(MouseEvent me) {}
   
}
  // end EditActivator

/**
  *  Inner CommandActivator watches over compilation and 
  *  running.  The user may set fluents at any time.  
  *  The file, class and classpath fluents are initially 
  *  calculated and recalculated whenever there is either a file
  *  or class spec change in the editor.  The user may set all fluent 
  *  values any time the compile or run dialog is visible; settings 
  *  are remembered until user changes them 
  *  or they are recalculated.
  *  
  */
class CommandActivator implements ActionListener { 
   String runcommand = "" ;           // from properties or user set
   String runPYcommand = "" ;         // from properties or user set
   String compilecommand = "" ;       // from properties or user set
   String lineparams = "" ;           // user may set
   String otherCompileSwitches = "" ; // user may set
   String otherRunSwitches = "" ;     // user may set
   String class_Path_Switch = "" ;    // from classPath or user set
   // Trick -> 
   String file_Spec ;                 // from fileSpec or user set
   String previous_file_Spec ;        // test for change by editor 
   String class_Spec ;                // from classSpec or user set
   String previous_class_Spec ;       // test for change by Java Editor 
   Vector<Process> processes ;       // store processes in case need to kill em

   CommandActivator() { 
      try { // not for AutoLogEditor
         //runcommand = settings.getProperty("runcommand").trim() ; 
         //runPYcommand = settings.getProperty("runPYcommand").trim() ; 
         //compilecommand = settings.getProperty("compilecommand").trim() ; 
      }
      catch(Exception x) { 
         System.out.println(x) ; 
      }
      processes = new Vector<Process>() ; 
   }

   public void actionPerformed(ActionEvent e) { 
      String command = e.getActionCommand() ;   
      if (command.equals("compile")) { // detect whether there is «..» embedded code
         if (autologGUI == null) createAutoLogGUI() ; 
         autologGUI.setVisible(true) ; // make sure viewer is visible
         // the source
         String src = ta.getText() + "\n" ; 
         if (src.indexOf("«")==-1)  {   // no «..» embedded code
            InputStream in = new ByteArrayInputStream(src.getBytes(Charset.forName("UTF-8"))) ; 
            autologGUI.setTheoryFromEditor(in) ;
         }
         else {  // extract embedded «..» code
            StringBuffer code_buffer = new StringBuffer() ;  // only the «code» pieces
            int begin = 0, 
                end = 0 ; 
            while (begin != -1 && end != -1) {  // begin AND end present
               begin = src.indexOf("«", begin) ;
               end = src.indexOf("»", begin) ;  
               if (begin!=-1 && end!=-1) {  // begin AND end STILL present
                  //System.out.println(src.substring(begin,end)) ;
                  code_buffer.append(src.substring(begin+1,end)) ; 
                  begin = src.indexOf("«",end+1) ;  // ... next «
               }
            }
            String code = code_buffer.toString() ;
            if (code.length()!=-1) {
               // create InputStream from code
               InputStream in = new ByteArrayInputStream(code.getBytes(Charset.forName("UTF-8"))) ; 
               autologGUI.setTheoryFromEditor(in) ;  // the extracted code
               //autologGUI.consoleArea.setText("#threads="+String.valueOf(Runtime.getRuntime().availableProcessors())+"\n") ;
            } // otherwise ignore... 
         }  // extract «..» code 
      }
      else if (command.equals("info")) {  
         // INFOFRAME
         try { 
            Class clz = activatorFor.getClass() ; 
            URL info = clz.getResource("Info.html") ;
            JEditorPane infoPane = 
               new JEditorPane(info) ; 
            infoPane.setBackground(backgroundColor) ;
            infoPane.setBorder(BorderFactory.createMatteBorder(8,12,0,8,Color.WHITE));
            JScrollPane scroller = new JScrollPane(infoPane) ; 
            JFrame infoFrame = new JFrame() ; 
            infoFrame.setTitle("AutoLog Version "+version) ; 
            infoFrame.getContentPane().add(scroller) ; 
            infoFrame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE) ; 
            infoFrame.setSize(400,650) ; 
            infoFrame.setVisible(true) ;                
         } catch (Exception xx) { System.out.println(xx) ; } 
      }
      else if (command.equals("symbols")) {  
         // INFOFRAME
         try { 
            Class clz = activatorFor.getClass() ; 
            URL symbols = clz.getResource("symbols.txt") ;
            JEditorPane infoPane = 
               new JEditorPane(symbols) ; 
            infoPane.setBackground(backgroundColor) ; 
            infoPane.setFont(new Font("monospaced",Font.PLAIN,24)) ;
            infoPane.setBorder(BorderFactory.createMatteBorder(8,8,0,3,Color.GRAY));
            JScrollPane scroller = new JScrollPane(infoPane) ; 
            JFrame infoFrame = new JFrame() ; 
            infoFrame.setTitle("Unicode Symbol Tables") ; 
            infoFrame.getContentPane().add(scroller) ; 
            infoFrame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE) ; 
            infoFrame.setSize(620,650) ;   // 600,650
            infoFrame.setVisible(true) ;                
         } catch (Exception xx) { System.out.println(xx) ; } 

      }
      else if (command.equals("circle")) {
         ta.replaceSelection("∘") ; 
      }
      else if (command.equals("ccircle")) {
         ta.replaceSelection("∙") ; 
      }
      else if (command.equals("cross")) {
         ta.replaceSelection("⨯") ; 
      }
      else if (command.equals("meet")) {
         ta.replaceSelection("∧") ; 
      }
      else if (command.equals("join")) {
         ta.replaceSelection("∨") ; 
      }
      else if (command.equals("not")) {
         ta.replaceSelection("¬") ; 
      }
      else if (command.equals("if")) {
         ta.replaceSelection("→") ; 
      }
      else if (command.equals("fi")) {
         ta.replaceSelection("←") ; 
      }
      else if (command.equals("iff")) {
         ta.replaceSelection("↔") ; 
      }
      else if (command.equals("all")) {
         ta.replaceSelection("∀") ; 
      }
      else if (command.equals("some")) {
         ta.replaceSelection("∃") ; 
      }
      else if (command.equals("top")) {
         ta.replaceSelection("⊤") ; 
      }
      else if (command.equals("bottom")) {
         ta.replaceSelection("⊥") ; 
      }
      else if (command.equals("intersect")) {
         ta.replaceSelection("∩") ; 
      }
      else if (command.equals("union")) {
         ta.replaceSelection("∪") ; 
      }
      else if (command.equals("member")) {
         ta.replaceSelection("∈") ; 
      }
      else if (command.equals("empty")) {
         ta.replaceSelection("∅") ; 
      }
      else if (command.equals("lambda")) {
         ta.replaceSelection("λ") ; 
      }
      else if (command.equals("pi")) {
         ta.replaceSelection("Π") ; 
      }
      else if (command.equals("sigma")) {
         ta.replaceSelection("Σ") ; 
      }
      else if (command.equals("possible")) {
         ta.replaceSelection("◇") ; 
      }
      else if (command.equals("necessary")) {
         ta.replaceSelection("□") ; 
      }
   }
}
  // end CommandActivator


/**
  *  PrintActivator controls the activities of printing
  *  and print preview.  The actual Printable is editor's 
  *  inner class JeditorPrinter.
  */
class PrintActivator implements ActionListener { 
   public void actionPerformed(ActionEvent e) { 
      String command = e.getActionCommand() ; 
      if (command.equals("pagesetup")) { 
         PrinterJob job = PrinterJob.getPrinterJob() ; 
         pageformat = job.pageDialog(pageformat) ; 
      }
      else if (command.equals("print"))  { 
         // Do not print an empty page.
         if (ta.getText().equals("")) return  ; 
         Graphics g = ta.getGraphics() ; 
         JeditorPrinter jedprinter = new JeditorPrinter() ; 
         int numpages = jedprinter.totalNumberOfPages(g,pageformat) ; 
         Book book = new Book() ; 
         book.append(jedprinter, pageformat, numpages) ; 
         PrinterJob job = PrinterJob.getPrinterJob() ; 
         job.setPageable(book) ; 
         if (job.printDialog()) { 
            try { 
               job.print() ;  
            } 
            catch(Exception x) { System.out.println(x) ; }
         }
      }
   }
}
  // end PrintActivator

} // end AutoLogEditor 


