package sam.machine.active ; 
import sam.machine.lang.* ; 

/**
  * abstract super class ActiveInference for extends by ActiveRule
  * and ActiveRewrite, which are two active inference components.
  * @version 10/21/2023  
  * The ActiveTheory class extends Vector<ActiveInference> 
  * which may have many ActiveRule(s) and ActiveRewrite(s).
  * See ActiveTheory.java to learn about how static and dynamic 
  * inferences are kept separate.
   */
public abstract class ActiveInference implements Runnable {                                                                                  
   /**
     * The start-1 index of a Rule or static Rewrite 
     * in the input autolog theory.
     */
   public int index ;    
}

/*   -----  NOTES  --------------------
  Example of dynamic ActiveRewrite & 'index'

  An Inference is the abstract super class for sam.machine.lang.Rule
  or sam.machine.lang.Rewrite.  An autolog source program is a 
  Vector<Inference>.

  An ActiveInference is the abstract super class for either a
  sam.machine.active.ActiveRule or sam.machine.active.ActiveRewrite.
  The ActiveInferences corresponding to source program Inferences
  are referred to as being "static". 

  There is also a dynamic ActiveRewrite which is constructed and added to the 
  the ActiveTheory after the BRANCH assertion of a new equality fact by 
  some ActiveRule.  A "static" ActiveRewrite has this.dynamic==false.

  A. For example, consider the theory (rules 1, 2,2)
    #1   true => p(a), a=b. 
    #2   b=c.  // static rewrite
    #3   p(c) => goal.

  BRANCH[i]
  0   true 
        |  
  1    p(a)    // ActiveRule #1 asserts BRANCH[1]
        |
  2    a=b     // ActiveRule #1 asserts eq fact BRANCH[2]
        |
  3    p(b)    // dynamic ActiveRewrite generated by eq fact at BRANCH[2] 
        |      // applied to fact at BRANCH[1]  
        |
  4    p(c)    // static ActiveRewrite #2 applied at BRANCH[3]
        |
  5    goal    //  ActiveRule #3 satisfyis antecedent at BRANCH[4] and asserts BRANCH[5]

  B. This example shows the need to withdraw a dynamic ActiveRewrite
     from an active theory on a new branch.

     true => p(a) | p(b)
     p(a)=> b=c.
     p(c) => goal.
     b=c =>  p(c).

  0     true 
        / \
  1  p(a)  p(b)
       |     
  2   b=c  
       |
  3   p(c)
       |
  4   goal
   
  Even though dynamic ActiveRewrite{b=c} lives below BRANCH[2] on the 
  left branch --- has instance variable -.atFact==2 --
  ActiveRewrite{b=c} dies when inference switches to the right branch 
  since 2 > 1(branch point), i.e. b=c is not in evidence on the right branch.  

  See ADN§6.11 regarding more dynamic ActiveRewrite methodology,
  inference descriptions and restrictions. 

  Autolog Proofs will report active rewrite actions in a similar manner.

-------------------------*/
