
package sam.machine.active ;
import sam.machine.lang.* ;
import java.util.* ;
import java.util.concurrent.* ; 
// can associate with AutoLog Viewer 
import sam.gui.search.* ; 

/*
  This file is an archive copy of active/ActiveTree.java
  as it was Tuesday, 8/29/2023 at 7:28 PM.
  This note is the only modification of the original
  file. The original file will be modified IN SITU.
*/

/**
  *  ActiveTree  -- See END NOTES
  *  Skolem Machine active inference management 
  *  (compare with colog14I/sam/infer/Tree )
  *  Many components are public to enable/promote 
  *  action spying for development code. 
  *  sam.machine.lang.Theory defines Analyzer data components.
  *  @version 8/23/2023 
  *    ActiveTree, ActiveTheory, ActiveRule, ActiveRewrite 
  *    all SHARE instance variable theory, and thus can
  *    access AA data via reference "this.theory". 
   */
public class ActiveTree implements Runnable {
   public Theory theory ;             // user theory/program
   public boolean predicative ;       // analyzer key profiles
   public ActiveTheory activetheory ; // dynamic 
   public Functor[] BRANCH ;
   public int depth ;       // restrict depth of BRANCH
   public int width ;       // restrict width of tree 
   public int branchNext ;  // next open BRANCH index   
   // Depth limit reached when branchNext == BRANCH.length.
   public int lshift = 1 ;  // default atpHashCode left shift
   public boolean log ;     // logging flag 
   ///////////////////////////////////////////////////////////
   //  Record branch fact locations in FactMap using either  
   //     predicative (colog14I) approach, or
   //     indexical (saturation) approach
   public ConcurrentHashMap<Integer,Vector<Integer>> FactMap ;  
   ///////////////////////////////////////////////////////////
   // inferenceServer processes staged ActiveRule and ActiveRewrite components
   public ExecutorService inferenceServer ;  

   ////////////////////
   // key profile source, MAX of 10 args 
   protected boolean[][] KEY = new boolean[1024][10] ; 
   // power=2^arity  arity: 0 1 2 3  4  5  6  7   8   9   10
   //    accelerator for pre-computation of KEY.
   protected static int[] power = {1,2,4,8,16,32,64,128,256,512,1024} ;  
   // call saturation setKEY for an indexical but not predicative theory
   protected void setKEY() {   
      // row 0 is all false
      for (int row=1 ; row < 1024 ; row++)  {
         // increment the key pattern
         boolean carry = true ;  // increment last row
         for (int col=0; col < 10; col++) {
            if (!KEY[row-1][col]) {  // previous row key
               if (carry) {
                  KEY[row][col]=true ;
                  carry=false ;
               }  
               else KEY[row][col]=false ;
            }
            else { // KEY[row-1][col]==true
               if (carry) {  
                  KEY[row][col]=false ;
                  carry=true ;
               } 
               else KEY[row][col]=true ;
            }
         }
      }
   }

   /**
    * Construct an ActiveTree if theory is indexical.
     */
   public ActiveTree(Theory thry, 
                     int deep, int wide,
                     boolean logit) { 
      this.theory = thry ; 
      if (this.theory.isIndexical())  { // construction 
         this.predicative = theory.isPredicative() ; // NOTE BENE
         int thry_size = thry.size() ; 
         if (thry_size > deep) this.depth=thry_size ;
         else this.depth=deep ; // avoid too shallow
         this.width = wide ; 
         this.log = logit ;
         this.FactMap = new ConcurrentHashMap<Integer,Vector<Integer>>() ; 
         this.BRANCH = new Functor[depth+1] ;  
         this.BRANCH[0] = new Functor("true",null,null,(byte)0) ; 
         this.branchNext = 1 ; 
         // can substitute many other Executors 
         this.inferenceServer = Executors.newWorkStealingPool() ; 
         // see sam.machine.engine.ActiveMod for some dev testing motivations
         this.activetheory = new ActiveTheory(this) ;  
      } // constructed ONLY if indexical
   }  
   /**
    * Create logging data structures. 
    * AutoLog Viewer always turns on logging.
     */
   public void setLog(boolean logit) { 
       this.log = logit ; 
   }

   /**                   
   *   atpHashCode calculations for a GROUNDED branch fact
   *   using a particular bound key pattern. 
   *   N.B., args are Functor terms.
   *   Use KEY list as key accelerator in rememberAll.
   *   ---------------------------------------------------
   *   Primitive recursive aritmetization
   *   for algebraic logic terms -- PRAALT
    */
   public int atpHashCode(Functor f, boolean[] key) {
      int h ; 
      if (f.fnctr != null)   // f parametric
         h = atpHashCode(f.fnctr) ;
      else h = f.index ;     // f named
      for (int i = 0 ; i < f.arity ; i++) // use key pattern
         h = (h << lshift) + (key[i] ? atpHashCode(f.args[i]) : 0) ; 
      return h ;        //   ^ N.B. NOT f.key[i] 
   }
   /** For embedded ground terms  */
   public int atpHashCode(Term t) { 
      Functor f = (Functor)t ;  
      int hc ; 
      if (f.fnctr != null)   // f parametric
         hc = atpHashCode(f.fnctr) ;
      else hc = f.index ;     // f named
      for (int i = 0 ; i < f.arity ; i++) { // accelerate?
         hc = (hc << lshift) + atpHashCode(f.args[i]) ;
      } 
      return hc ; 
   }

   /** 
     *  Remember a bound facts's branch location (bloc) 
     *  in FactMap. This method remembers All possible key 
     *  profiles using KEY (relevant to fact.arity).    
     *  Separate method remember uses only analysed key profiles, 
     *  where appropriate (e.g. for predicative theories).  
      */
   protected void rememberAll(Functor fact, int bloc) { 
      int arity = fact.arity ; 
      if (arity == 0) { // a named predicate or constant arg
         Vector<Integer> locations = FactMap.get(fact.index) ; 
         if (locations == null) {
            Vector<Integer> vec = new Vector<Integer>() ;
            vec.add(bloc) ;
            FactMap.put(fact.index, vec) ; 
         }
         else if (!locations.contains(bloc))
               locations.add(bloc) ; 
         else {} // bloc already recorded
         return ; 
      }
      else { // arity ≥ 1  so assert to FactMap by key 
         for (int row=0 ; row < this.power[arity] ; row++) {    
            // ✪ USE ALL OF THE key profiles RELEVANT TO arity ...
            boolean[] key = this.KEY[row] ;
            int hcode = atpHashCode(fact,key) ; 
            Vector<Integer> locations = FactMap.get(hcode) ;
            if (locations == null) { 
               Vector<Integer> vec = new Vector<Integer>() ;
               vec.add(bloc) ;
               FactMap.put(hcode, vec) ; 
            }
            else if (!locations.contains(bloc)) { 
               locations.add(bloc) ;  
            }
            else {} // bloc already recorded
         }
      }
   }   
 
   /**
     * Start the prover. E.g., the "run" command in the AutoLog
     * Viewer initiates the following run method. Staged 
     * inference will use inferenceServer.invokeAll(definite) 
     * method on the definite ActiveTheory inference components
     * and then will use inferenceServer.invoke(e) on a chosen
     * existential ActiveTheory inference component (if any).
      */
   public void run() { 
      // UNDER CONSTRUCTION ...
   }

   /**
     * Short debug tests...
     * Examples illustrating index's and some indexing details.
     * If !thy.isIndexical() then exceptions would occur.
      */
   public static void main(String[] args) { 
      // Provide Theory for inspections...
      String thy =     // test theory input
         "a=b. true=>p(Z). f(X)=X. true=>c=d. f(a)=g(b). p(X)=>q(X,Z)." ;
      Theory theory = Theory.translate(thy) ; 
      AutoLogAnalyzer.analyze(theory) ;
      System.out.println(" Short debug tests...") ; 
      System.out.println(" Examples illustrating index's and indexing details. \n") ; 
     ///// A.
      System.out.println("A. theory Vector<Inference> is indexed by AA:line#32 ") ; 
      //System.out.println(theory) ; // display .index and then inference
      Iterator it = theory.iterator() ;
      while (it.hasNext()) { // show .index and Inference
         Inference infrnc = (Inference)(it.next()) ;
            System.out.println(" "+infrnc.index +" "+ infrnc) ;
      }  System.out.print("\n") ; 

     ///// B.
      System.out.println("B. Rewrite vs. eq-Functor ...") ; 
      // check index of first Rewrite, should be 3)
      System.out.print(" Rewrite@0: "+ theory.elementAt(0)) ; // &
        System.out.println("  .index = "+theory.elementAt(0).index) ; // 0
      Rewrite rw = (Rewrite)theory.elementAt(0) ;
      System.out.println(" Rewrite@0 is bound: U="+rw.U+", "+"V="+rw.V) ; // √
      System.out.print(" Rewrite@2: "+ theory.elementAt(2)) ; // &
        System.out.println("  .index = "+theory.elementAt(2).index) ;   
      System.out.print(" Rewrite@4: "+ theory.elementAt(4)) ; // &
        System.out.println("  .index = "+theory.elementAt(4).index) ;   
      System.out.println(" theory.name[3]: "+ theory.name[3]) ; 
      System.out.println(" theory.name[4]: "+ theory.name[4]) ; 
      Rule r = (Rule)theory.elementAt(3) ;
      System.out.println("Rule@3: "+r) ; 
      Functor literal = (Functor)(r.consequent.elementAt(0).elementAt(0)) ;
      System.out.println("Functor «c=d».index in Rule@3 = " + literal.index) ; // 3
      System.out.print("\n") ; 

     ///// C
      System.out.println("C. tree.activetheory generates BRANCH facts for grounded Rewrite(s) ... ") ; 
      ActiveTree tree = 
         new ActiveTree(theory,10,10,false) ; // no log for test
      System.out.println("tree.BRANCH.length = "+tree.BRANCH.length) ; // 11 √
      System.out.println("tree.branchNext = " + tree.branchNext ) ; //
      System.out.println("BRANCH[0] : " + tree.BRANCH[0]) ; // √
      System.out.println("BRANCH[1] : " + tree.BRANCH[1]) ; // √
      System.out.println("BRANCH[2] : " + tree.BRANCH[2]) ; // √
      System.out.println("tree.BRANCH[0].index = "+tree.BRANCH[0].index) ;  // √ 3 fixed 
      System.out.println("tree.BRANCH[1].index = "+tree.BRANCH[1].index) ;  // √ 3 fixed 
      System.out.println("tree.BRANCH[2].index = "+tree.BRANCH[2].index) ;  // √ 3 fixed 

     ///// D report() generated by tree.activetheory ...
      System.out.println("\n D. report() generated by tree.activetheory re definite & existential") ; 
      System.out.println(tree.activetheory.report()) ; 
   } // end main
} // end ActiveTree

/* END NOTES 
  -★ Each ActiveInference instance (rule or rewrite)
   should specify its own remember methods, in order 
   to facilitate concurrent inferences.  

  -★ Develop a new remember method for indexical
    (non predicative) theories. See 
    active_dev/activeDesignScenarios/★ActiveTree.txt

Monday 5/28/2023  inspect code
"FactMap"
33
36
86
"rememberALL"> to be deprecated see active_dev/ActiveTree[Revise].txt
108               
137
145
149
156
160
164 <<<<< LAST "rememberAll", "FactMap"
∴ "rememberAll"(...), "FactMap" not used in main tests
<<<


*/

