import java.util.* ;
import java.util.concurrent.* ;
import java.time.Duration;
import java.time.Instant;

/**
  *  Runnable concurrency execution test example. 
  *  The runnable task stores its data computation (sum),
  *  which can be accessed after processing.
  *  @version 12/31/2022
   */
public class RunMe implements Runnable { 
    public int n ; 
    public int sum  ;

    public RunMe(int n) {
       this.n = n ;
    }

   public void run() {
      int s = 0 ;
      for(int i=1; i<=this.n ; i++)
         s+=i ;
      this.sum = s ;
   }

   ///////////////////////////////////////////////
   /////  Concurrency Executors examples 
   /////  Boilerplate code
   ///////////////////////////////////////////////
   /**  MAIN
     * Compare   Executors.newFixedThreadPool(-) 
     *      vs   Executors.newVirtualThreadPerTaskExecutor()
     * running concurrent list of RunMe tasks
      */
   public static void main(String[] args) { 
      ///// create list of independent RunMe tasks
      int N = 100000 ; // number(edit) of RunMe tasks 
      Vector<RunMe> tasks = new Vector<RunMe>() ;
      for (int n=0 ; n<N ; n++)
         tasks.add(new RunMe(50000))  ; 
      System.out.println(tasks.size()+" tasks") ;
  
      ///// construct executor, CHOOSE: 
//      ExecutorService executor = Executors.newWorkStealingPool() ; 
//      ExecutorService executor = Executors.newCachedThreadPool() ; 
//      ExecutorService executor = Executors.newThreadPerTaskExecutor(Executors.defaultThreadFactory()) ; 
      ExecutorService executor = Executors.newFixedThreadPool(18) ; 
//      ExecutorService executor = Executors.newVirtualThreadPerTaskExecutor() ; 
//             --release 19 --enable-preview

      ///// submit all tasks to executor 
      Instant start = Instant.now();
      try{ 
         for(int n=0 ; n<N ; n++) 
            executor.submit((RunMe)(tasks.elementAt(n))) ;
      } catch(Exception e) { System.out.println(e) ; }
      
      ///// shutdown and await termination
      executor.shutdown() ;    // needs to finish, so
      try {          // added for virtual thread version
           if (!executor.awaitTermination(60, TimeUnit.SECONDS)) {
               executor.shutdownNow();
          }
      } catch (InterruptedException ex) {
           executor.shutdownNow();
           Thread.currentThread().interrupt();
      }
      
      ///// report 
      Instant finish = Instant.now();
      System.out.println("shutdown " + executor.isShutdown()) ;
      System.out.println("terminated " +executor.isTerminated()) ;
      long timeElapsed = Duration.between(start, finish).toMillis();  //in millis
      System.out.println("Total elapsed time : " + timeElapsed+"ms");
      ///// spy on tasks
      System.out.println(tasks.elementAt(9050).sum) ;
       
   }
}
    

   
/* 
I suspect that this example does not demonstrate any definite
superiority for virtual threads. It does demonstrate how too many
fixed threads can slow things down.

iMPro (18 core)
            N     time
work steal 100K   236ms    ok
cached     100K   461ms    X
threadPerT 100K   5548ms   X  (errors)
fixed(18)  100K   160ms |  ok  <<<<
virtual    100K   215ms |  ok

Good times indeed for both fixed and virtual.
However, the task is PRA.
*/
