//package com.howtodoinjava.loom;

import java.util.Random;
import java.time.Duration;
import java.time.Instant;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.Executors;
import java.util.concurrent.Callable;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
// --release 19 --enable-preview

public class HelloLoom {
  public static void main(String[] args) throws InterruptedException {

    final AtomicInteger atomicInteger = new AtomicInteger();

    Runnable runnable = () -> {
      try {
        Thread.sleep(Duration.ofSeconds(1));
      } catch(Exception e) {
          System.out.println(e);
      }
      System.out.println("Work Done - " + atomicInteger.incrementAndGet());
      //System.out.println("Work Done - " + atomicInteger.getAndIncrement());
    };

    Instant start = Instant.now();

    //CHOOSE ONE: 
    //try (var executor = Executors.newFixedThreadPool(100)) { //A
    try (var executor = Executors.newVirtualThreadPerTaskExecutor()) { //B
      for(int i = 0; i < 10_000; i++) {
        executor.submit(runnable);
      }
    }

    Instant finish = Instant.now();
    long timeElapsed = Duration.between(start, finish).toMillis();  //in millis
    System.out.println("Total elapsed time : " + timeElapsed);
  }
}

// N.B. public static ExecutorService newVirtualThreadPerTaskExecutor()

// javac switches:  --release 19 --enable-preview   
// java  switches:   --enable-preview                
//                         Air      iMP     MBP
// # cores                 2        18      6
//A fixed elapsed time :   100419   100246  100408
//B virtual elapsed time : 2750     1173    1469
//  # system cores seems  less relevant (so far)!                                           

// Use 18 (rather than 100) threads on iMP
//A fixed elapsed time:              557935   (~5.6x MORE elapsed time)
// Use 200 (rather than 100) threads on iMP
//A fixed elapsed time:              50148   (~1/2 the elapsed time)
