////////////////
// Loom2.java   Callable version of Runnable HelloLoom
// >>>  newVirtualThreadPerTaskExecutor().invokeAll(Vector<Callable> tasks) ...
////////////////

import java.util.Iterator ;
import java.util.List ;
import java.util.Vector ;
import java.util.Random;
import java.time.Duration;
import java.time.Instant;
import java.util.concurrent.Future ;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.Executors;
import java.util.concurrent.Callable;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
// --release 19 --enable-preview

public class Loom2 {

   public static void main(String[] args) throws InterruptedException {

      final AtomicInteger atomicInteger = new AtomicInteger(); 
      final Random random = new Random() ; // return for callable
         // introduces some random delay

      Callable<Integer> task = () -> {
        try {
          Thread.sleep(Duration.ofSeconds(1));  
        } catch(Exception e) {
            System.out.println(e);
        }
        int retval =  random.nextInt(10) ; // 0...9 
        System.out.println("Work Done - " + atomicInteger.incrementAndGet());
        return retval ; 
      } ;  

      List<Callable<Integer>> tasklist = new Vector<Callable<Integer>>() ;
      for(int i = 0; i < 10_000; i++) 
        tasklist.add(task) ;
   
      List<Future<Integer>> resultList = null;

      Instant start = Instant.now();
     
      // public static ExecutorService newVirtualThreadPerTaskExecutor()
      ExecutorService executor = 
         Executors.newVirtualThreadPerTaskExecutor() ;

//////// Difficult read for the spec               
//    For each class T, method invokeAll(...) applied to a
//    Collection extending Callable<T> returns a  List<Future<T>>
//    and throws InterruptedException
// public <T> List<Future<T>> invokeAll(Collection<? extends Callable<T>> tasks)
////////                        throws InterruptedException
      try {
         resultList = executor.invokeAll(tasklist) ; 
      } catch (InterruptedException e) {
           System.out.println(e) ;  
      }
         
      executor.shutdown() ; 
      Instant finish = Instant.now();
      long timeElapsed = Duration.between(start, finish).toMillis();  //in millis
      System.out.println("Total elapsed time : " + timeElapsed);

      // Show resultList
      try {// catch InterruptedException, ExecutionException 
         StringBuffer sbuf = new StringBuffer() ; 
         Iterator it = resultList.iterator() ; 
         while (it.hasNext()) {
            sbuf.append(((Future)it.next()).get()) ; // get int from Future
         }
         System.out.println(sbuf.toString()) ;
       } catch (Exception e) { // InterruptedException, ExecutionException
          System.out.println(e) ;
       }
       
   } // end main

} // end Loom2

/*    1s     10s    20s      delay--line 31
MBAir 2238   11478  22546    2 cores
iMP   1208   10181  20198    18 cores

How can 10K threads waiting 20 seconds finish
in 10-12 seconds?

The MBAir vs iMP times seem NOT to make sense !? 
Code line 31 does NOT "stall" a Thread for a Duration. 

*/
