package sam.machine.lang ; 

import java.util.* ;
import java.io.* ; 
import sam.util.* ;

/**    
  *  Analyze AutoLog theory.  
  *  @version 6/17/2023 share indexmap typemap  with theory, line #116
  *  Rule keyPatterns and Lterm static rewriteKeyPatterns 
  *  separated.
   */
public class AutoLogAnalyzer { 
   // non-generic static renaming, can use for [] and casting, w/o warnings
   static class VarMap  extends HashMap<Variable,Integer> {} 
   static class VarSet  extends HashSet<Variable> {} 
   static class PatternVector  extends Vector<Pattern> {}  
             // ^ used to create generic arrays of Vector<Pattern> at * below
   private static String NAS ="/" ;   // name arity separator 

   /**
     *  Analyze AutoLog theory,
     *  call analyze(Theory).
      */
   public static void analyze(Theory theory) {       
      trimToSize(theory) ;     // trim all vectors-WHEN?
      indexFunctors(theory) ; // Both Rule and Rewrite 
      Iterator it = theory.iterator() ; 
      int indx = 0 ;   
      while (it.hasNext())  {
         Inference inf = (Inference)it.next() ;
         inf.index = indx ;  // index the inference
         if (inf instanceof Rule) {
            Index[] forward = indexRuleVariables((Rule)inf) ;  // Vector<Integer>[]
            ((Rule)inf).forward = forward ;
            setOrder((Rule)inf) ; 
            ((Rule)inf).complex = complex((Rule)inf) ; 
         }
         else if (inf instanceof Rewrite) {
            indexRewriteVariables((Rewrite)inf) ;  
            setOrder((Rewrite)inf) ; // 
         }
         indx++ ; 
      } 
      // Compute Rewrite converses  
      for (int i = 0 ; i < theory.size()-1 ; i++) { 
         if (theory.elementAt(i) instanceof Rewrite) { 
            Rewrite r1 = (Rewrite)(theory.elementAt(i)) ;
            if (theory.elementAt(i+1) instanceof Rewrite) { 
               Rewrite r2 = (Rewrite)(theory.elementAt(i+1)) ;
               if (r1.converse(r2)) { 
                  r1.converse = r2 ; 
                  r1.converse_index = r2.index ;
                  r2.converse = r1 ;
                  r2.converse_index = r1.index ;  
                  i += 2 ; // skip over i+1 because found converse pair            
               } // second not a converse
            } // second not a rewrite  
         } // first not a rewrite
      }
      setKeyPatterns(theory) ;  // NEW 2/8/2022 -- literal or rewrite key patterns
   }

///////////////////////////////
///// BEGIN  INDEX FUNCTORS 
///////////////////////////////
   /**
     *  Computes index field for functors. 
     *  Call after all rules are added to theory.  
     *  Constructs name[] and type[] for theory.
      */
   protected static void indexFunctors(Theory theory) {
      // functor name/arity --> index 
      HashMap<String,Integer> indexmap = new HashMap<String,Integer>() ; 
      indexmap.put("goal"+NAS+"0",Theory.GOAL_INDEX) ;    // 1
      indexmap.put("false"+NAS+"0",Theory.FALSE_INDEX) ;  // 2
      indexmap.put("true"+NAS+"0",Theory.TRUE_INDEX) ;    // 0
      indexmap.put("="+NAS+"2",Theory.EQ_INDEX) ;         // 3
      // functor name/arity --> type 
      HashMap<String,Byte> typemap = new HashMap<String,Byte>() ;
      typemap.put("goal"+NAS+"0",Functor.FUNCTIONAL) ; 
      typemap.put("false"+NAS+"0",Functor.FUNCTIONAL) ; 
      typemap.put("true"+NAS+"0",Functor.FUNCTIONAL) ; 
      typemap.put("="+NAS+"2",Functor.INFIX) ; 
      Iterator it = theory.iterator() ; 
      theory.count = 4 ; 
      while (it.hasNext()) { 
         Inference inf = (Inference)it.next() ; 
         if (inf instanceof Rule) {
            Conjunction a = ((Rule)inf).antecedent ; 
            Iterator ait = a.iterator() ; 
            while(ait.hasNext()) { 
               Functor f = (Functor)ait.next() ; 
               indexFunctor(f,theory,indexmap,typemap) ;             
            }
            Disjunction d = ((Rule)inf).consequent ; 
            Iterator dit = d.iterator() ; 
            while (dit.hasNext()) { 
               Conjunction c = (Conjunction)dit.next() ;
               Iterator cit = c.iterator() ; 
               while (cit.hasNext()) { 
                  Functor f = (Functor)cit.next() ; 
                  indexFunctor(f,theory,indexmap,typemap) ; 
               }
            }
         }
         else if (inf instanceof Rewrite) { // should be Rewrite A->B
            Rewrite m = (Rewrite)inf ; 
            Term[] args = new Term[2] ;
            args[0] = m.Lterm ; args[1] = m.Rterm ;
            Functor f = new Functor("=", null,args,(byte)2) ; 
            indexFunctor(new Functor("=", null,args,(byte)2), theory,indexmap,typemap) ;
         }
         else { }  // nothing yet
      }
      // share indexmap and typemap with theory so that
      theory.indexmap = indexmap ; 
      theory.typemap = typemap ; 
      // now all functors are mapped, so create 
      // theory  name[], arity[] ,type[]
      theory.name = new String[indexmap.size()] ; 
      theory.arity = new int[indexmap.size()] ; 
      theory.type = new byte[indexmap.size()] ; 
      Iterator fcns = indexmap.keySet().iterator() ; 
      while (fcns.hasNext()) { 
         String s = (String)fcns.next() ; 
         int i = indexmap.get(s) ; 
         byte t = typemap.get(s) ;
         int k = s.indexOf(NAS) ; 
         String n = s. substring(0,k) ; 
         theory.name[i] = n ; 
         theory.arity[i] = Integer.valueOf(s.substring(k+1,s.length())) ; 
         theory.type[i] = t ;
      }
   }

   // map indexes for functor and arguments
   protected static void indexFunctor(Functor f,Theory thy,HashMap<String,Integer> imap,HashMap<String,Byte> tmap) { 
      Object probe = imap.get(f.name+NAS+f.arity) ; 
      if (probe != null) 
         f.index = (Integer)probe ; 
      else {
         f.index = thy.count ; // via name/arity
         imap.put(f.name+NAS+f.arity,thy.count) ; 
         thy.count++ ;
      }      
      tmap.put(f.name+NAS+f.arity,f.type) ;
      // analyze fnctr 
      if (f.fnctr != null && 
              f.fnctr instanceof sam.machine.lang.Functor)
          indexFunctor((Functor)f.fnctr,thy,imap,tmap) ; 
      // also do for functor args
      for (int i = 0 ; i < f.arity ; i++) {
         Term t = f.args[i] ; 
         if (t instanceof sam.machine.lang.Functor) 
            indexFunctor((Functor)t,thy,imap,tmap) ; 
      }
   }
///////////////////////////////
///// END  INDEX FUNCTORS 
///////////////////////////////

/////////////////////////////////////////////
///// BEGIN  INDEX RULE VARIABLES 
///// for all Functor
////     boolean[] key, are args bound?
////     in Functor int[] order, to match args
///      in Rule int[] forward,  consequents bound after literal factor matched
/////////////////////////////////////////////
   /**
     *  Index RULE variables.
     *  Automatically sets literal/predicate 'key' patterns 
     *  for uniform tableing, literal/predicate match 'order'
     *  and literal/predicate 'factual' field.
     *  @return the forward lists for r
      */
   protected static Index[] indexRuleVariables(Rule r) { 
      Index[] forward = new Index[r.antecedent.size()] ; 
      for (int j = 0 ; j < r.antecedent.size() ; j ++) 
         forward[j] = new Index() ; 
      VarMap Umap = new VarMap() ;
      VarSet[] Uvars = indexAntecedentVariables(r.antecedent,Umap) ; 
      r.U = Umap.size() ; 
      Disjunction d = r.consequent ; 
      int s = 0 ;
      VarMap tmp = new VarMap() ;
      for(int i = 0 ; i < d.size() ; i++) { 
         tmp.clear() ; 
         tmp.putAll(Umap) ;  // clone warns
         VarSet vs = indexConsequentVariables(d.elementAt(i),tmp,Umap.keySet().size()) ;  
         if (tmp.size() > s) s = tmp.size() ; 
         forward[forward(Uvars,vs)].add(i) ;
      }   
      r.V = s ; 
      r.existential = r.V - r.U ;  
      r.varindex = varindex(r)  ; 
      return forward ;
   }

   /**
     *  @return smallest i such that antsets[i] contains all 
     *  universal variables from dset.  e.g., 
     *     r=    p(X), p(Y), p(Z) => q(X), r(Y) | q(Z) | q(Y).
     *           0     1     2      <    0   >  < 1 >   < 2 >
     *         r.forward[0]=[]     
     *               r.forward[1]=[0,2]
     *                     r.forward[2]=[1]
      */
   protected static int forward(VarSet[] antsets,VarSet dset) { 
      int fwrd = 0 , max = antsets[antsets.length-1].size() ;  
      Iterator it = dset.iterator() ; 
      while(it.hasNext()) {
         Variable v = (Variable)it.next() ; 
         if (v.index < max) {
            for(int i = 0 ; i < antsets.length ; i++) { 
               if (antsets[i].contains(v)) { 
                  if (i > fwrd) fwrd = i ; 
                  break ; // smallest i for this v
               }
            }
         }
      }
      return fwrd ; 
   }  

   /**
     *  Index variables in ANTECEDENT conjunction 
     *  @return cumulative variable sets for factors, left-to-right
      */
   protected static VarSet[] indexAntecedentVariables(Conjunction c, VarMap map) { 
      VarSet[] set = new VarSet[c.size()] ;  
      for(int i = 0 ; i < c.size() ; i++) { 
         Functor f = (Functor)c.elementAt(i) ; 
         f.key = new boolean[f.arity] ;  // SIGNAL f is predicate 
         indexAntecedentVariables(f, map) ; 
         set[i] = new VarSet() ; 
         set[i].addAll(map.keySet()) ; 
      }
      return set ; 
   }

   /** 
     *  Index the variables inside ANTECEDENT terms.
     *  Automatically initializes and sets antecedent predicate key.
      */
   protected static void  indexAntecedentVariables(Term t, VarMap map) { 
      if (t instanceof Variable) { 
         Object ob = map.get((Variable)t) ;
         if (ob != null) ((Variable)t).index = (Integer)ob ;  
         else { 
            ((Variable)t).index = map.size() ; 
            map.put((Variable)t,map.size()) ;
         }
      }
      else if (t instanceof Functor) {
         VarMap save_map ;    // to check for new variables
         Functor f = (Functor)t ; 
         if (f.fnctr !=null) {
            save_map = (VarMap)map.clone() ;     // vars so far // 
            indexAntecedentVariables(f.fnctr, map) ;
            VarSet vs = new VarSet() ;                                 
            findVars(f.fnctr,vs) ;  // actual variables occur in fnctr
            f.predicative = vs.isEmpty() ;  
               // f predicative provided no variables occur in fnctr
               // predictive --> indexical
            f.indexical = (((VarMap)save_map).keySet()).containsAll(vs) ; 
               // f indexical provided all variables in fnctr occurred previously
         }
         save_map = (VarMap)map.clone() ;        // vars so far // 
         for(int i = 0 ; i < f.arity ; i++) {  
            indexAntecedentVariables(f.args[i], map) ;   
            if (f.key != null) { // f is a predicate, so set f.key[i]
               VarSet vs = new VarSet() ;                                 
               findVars(f.args[i],vs) ;                                   
               f.key[i] = (((VarMap)save_map).keySet()).containsAll(vs) ; // new var? //
            }
         }
      }
      else {}
   }   

   // Compute the VarSet for a term.   
   // Modified to include fnctr.                          
   private static void findVars(Term t, VarSet set) { 
      if (t instanceof Variable)  set.add((Variable)t) ; 
      else { 
         Functor f = (Functor)t ;
         if (f.fnctr != null) findVars(f.fnctr,set) ; 
         for (int i = 0 ; i < f.arity ; i++) 
            findVars(f.args[i],set) ; 
      }
   }

   /*
     *  Compute .varindex for Rule
     *  This method displays (checks) indexes  
     *  already computed for variables.  
     */
   private static String varindex(Rule r) {  
      StringBuffer sb = new StringBuffer() ; 
      varindex(r.antecedent, sb) ;
      sb.append(" => ") ; 
      Iterator consit = r.consequent.iterator() ; 
      while (consit.hasNext()) { 
        varindex((Conjunction)consit.next(), sb) ; 
        if (consit.hasNext()) sb.append(" | ") ;
      }
      return sb.toString() ;
   }

   // compute the varindex for a conjunction 
   private static void varindex(Conjunction c, StringBuffer sb) { 
      VarSet set = new VarSet() ;
      Iterator cit = c.iterator() ;
      while (cit.hasNext())  
         findVars((Term)cit.next(), set) ;  
      Iterator vit = set.iterator() ; 
      while (vit.hasNext()) { 
         Variable v = (Variable)vit.next() ; 
         sb.append(v.name+":"+v.index) ; 
         if (vit.hasNext()) sb.append(" ") ; 
      }    
   }

   /**
     *  Index variables in CONSEQUENT conjunction 
     *  @return set of universal variables in the conjunction
      */
   protected static VarSet indexConsequentVariables(Conjunction c,VarMap map,int U) { 
      VarSet set = new VarSet() ;  
      for(int i = 0 ; i < c.size() ; i++) { 
         Functor f = (Functor)c.elementAt(i) ; 
         f.key = new boolean[f.arity] ;  // !=null signals f is predicate, needs key
         set.addAll(indexConsequentVariables(f,map,U)) ; 
      }
      return set ; 
   }  

   /**
     *  Index variables in a CONSEQUENT term.
     *  map is index map of variables, to which existential variables are also put.
     *  U is the number of universal variables in the antecedent .
     *  @return set of universal variables in this consequent conjunction
      */
   protected static VarSet indexConsequentVariables(Term t,VarMap map, int U) { 
      VarSet set = new VarSet() ;
      if (t instanceof Variable) { 
         Object ob = map.get((Variable)t) ;
         if (ob != null) { 
            ((Variable)t).index = (Integer)ob ;  
            if ((Integer)ob < U) set.add((Variable)t) ; // a universal variable
         }
         else { 
            ((Variable)t).index = map.size() ; 
            map.put((Variable)t,map.size()) ;
         }
      }
      else if (t instanceof Functor) { 
         VarMap save_map ;    // to check for new variables
         Functor f = (Functor)t ; 
         if (f.fnctr !=null) { 
            save_map = (VarMap)map.clone() ;     // vars so far // 
            indexConsequentVariables(f.fnctr,map,U)  ;
            VarSet vs = new VarSet() ;                                 
            findVars(f.fnctr,vs) ;  // actual variables occur in fnctr
            f.predicative = vs.isEmpty() ;  
               // f predicative provided no variables occur in fnctr
               // predictive --> indexical
            f.indexical = (((VarMap)save_map).keySet()).containsAll(vs) ; 
               // f indexical provided all variables in fnctr occurred previously
         }
         save_map = (VarMap)map.clone() ;      // vars so far                   
         for(int i = 0 ; i < f.arity ; i++) {      
            VarSet uvars = indexConsequentVariables(f.args[i],map,U) ;
            set.addAll(uvars) ;          
            if (f.key != null) { // f is a predicate, so set f.key[i]
               VarSet vs = new VarSet() ;                                 
               findVars(f.args[i],vs) ;                                  
               f.key[i] = (((VarMap)save_map).keySet()).containsAll(vs) ; // new var ?
           }
         }
      }
      else {}
      return set ;
   }   
///////////////////////////////
///// END INDEX RULE VARIABLES 
///////////////////////////////
/////////////////////////////////////////////
///// BEGIN  INDEX MODULATOR VARIABLES 
/////    Need warning if Lterm is impredicative,
/////      -> detect  impredicate Lterm <-
/////////////////////////////////////////////
   /**
     *  Index Rewrite variables. 
     *  Sets .index of all variables in Lterm and Rterm
     *  If Lterm instanceof Functor calculate '.key' '.order', and '.factual'. 
     *  Calculate complexity of modulator -- Rterm more complex than Lterm.
     *  @version 5/12/2022
      */
   protected static void indexRewriteVariables(Rewrite m) {  
      if (m.Lterm instanceof Variable) m.warnings+= "WARNING: variable Lterm. " ; 
      VarMap map = new VarMap() ;
      /////
      if (m.Lterm instanceof Functor) { // create .key for top functor 
         Functor f = (Functor)m.Lterm ; 
         f.key = new boolean[f.arity] ; 
      }      
      /////
      indexRewriteVariables(m.Lterm,map) ; // set .key for Functor
      m.U = map.size() ; 
      if (map.isEmpty()) m.factual = true ; // factual≈no variables in Lterm
      indexRewriteVariables(m.Rterm,map) ; 
      m.V = map.size() ; 
      m.predicative= predicative(m) ;  // Lterm is Variable or free var in fnctr
      // if (!(m.Lterm instanceof Variable) && !m.predicative) m.warnings+= "WARNING: impredicative Lterm. " ; 
      m.existential = map.size() - m.U ;
      if (m.existential> 0) m.warnings+= "WARNING: free variable in Rterm. " ; 
      m.varindex = varindex(m)  ; 
      // Does modulator introduce more complex terms?
      m.complex = termComplexity(m.Rterm) > termComplexity(m.Lterm) ; // 1/3/2021
   }

   /**
     *  Index the variables in Lterm or Rterm of Rewrite, and 
     *      for a=true construct f.key for Functor f in Lterm - 
     *      f.key[i]=true is f.args[i] contains no variable.
      */
   protected static void indexRewriteVariables(Term t,VarMap map) { 
      if (t instanceof Variable) { 
         Object ob = map.get((Variable)t) ;
         if (ob != null) ((Variable)t).index = (Integer)ob ;  
         else { 
            ((Variable)t).index = map.size() ; 
            map.put((Variable)t,map.size()) ;
         }
      }
      else if (t instanceof Functor) {
         VarMap save_map ;    // to check for new variables
         Functor f = (Functor)t ; 
         if (f.fnctr !=null) {
            save_map = (VarMap)map.clone() ;     // vars so far // 
            indexRewriteVariables(f.fnctr, map) ;
            VarSet vs = new VarSet() ;                                 
            findVars(f.fnctr,vs) ;  // actual variables occur in fnctr
            f.predicative = vs.isEmpty() ;  
               // f predicative provided no variables occur in fnctr
               // predictive --> indexical
            f.indexical = (((VarMap)save_map).keySet()).containsAll(vs) ; 
               // f indexical provided all variables in fnctr occurred previously
         }  
         /////   & then f.args
         save_map = (VarMap)map.clone() ;   // vars so far // 
         for(int i = 0 ; i < f.arity ; i++) {  
            indexRewriteVariables(f.args[i], map) ;   // no .key
            if (f.key != null) { // f is a predicate, so set f.key[i]
               VarSet vs = new VarSet() ;                                 
               findVars(f.args[i],vs) ;                                   
               f.key[i] = (((VarMap)save_map).keySet()).containsAll(vs) ; // ¬new 
            }
         }
         /////  
      }
      //else {}
   }   

   /**
     *  Inspection method for the varindex of a rewrite.
      */ 
   protected static String varindex(Rewrite m) { 
      StringBuffer sb = new StringBuffer() ; 
      varindex(m.Lterm,sb)  ; 
      sb.append(" = ") ;
      varindex(m.Rterm,sb)  ; 
      return sb.toString() ;
   }

   // compute varindex for modulator terms
   private static void varindex(Term t, StringBuffer sb) { 
      VarSet set = new VarSet() ; 
      findVars(t,set) ;
      Iterator vit = set.iterator() ; 
      while (vit.hasNext()) { 
         Variable v = (Variable)vit.next() ; 
         sb.append(v.name+":"+v.index) ; 
         if (vit.hasNext()) sb.append(" ") ; 
      }    
   }

   // compute whether modulator is predicative -- naive early approximation
   // Use Rewrite.isPredicative for autolog22
   private static boolean predicative(Rewrite m) { 
      if (m.Lterm instanceof Variable) return false ;
      else {  // m.Lterm is a Functor
         Functor f = (Functor)(m.Lterm) ;
         if (f.fnctr != null)  { // find vars in f.fnctr ...
            VarSet set = new VarSet() ;
            findVars(f.fnctr,set) ;
            if (set.isEmpty()) return true ;
            else return false ;
         }
         else return true ;  // not functorial 
      } 
   }

/////////////////////////////////////////////
///// END  INDEX MODULATOR VARIABLES 
/////////////////////////////////////////////

///////////////////////////////
///// BEGIN  setOrder  
///// 
///////////////////////////////
   /**
     *  Use predicate key to set order and factual
      */
   protected static void setOrder(Rule r) { 
      r.antecedent.factual = true ;  // unless not ...
      Iterator ant = r.antecedent.iterator() ;  
      while (ant.hasNext()) {
         Functor f = (Functor)ant.next() ;
         f.order = order(f.key) ;
         f.factual = factual(f.key) ;
         if (!f.factual) r.antecedent.factual = false ;    
      }
      if (r.antecedent.factual) r.factual = true ;
      Iterator d = r.consequent.iterator() ; 
      while (d.hasNext()) {
         Conjunction conj = (Conjunction)d.next() ;
         conj.factual = true ;  // unless not ...
         Iterator c = conj.iterator() ; 
         while (c.hasNext()) { 
            Functor f = (Functor)c.next() ;
            f.order= order(f.key) ;
            f.factual = factual(f.key) ;  
            if (!f.factual) conj.factual = false ;
         }
      }
   }

   /**
     *  @return order int[] based on boolean[] key 
      */
   protected static int[] order(boolean[] key) { 
      int[] ord = new int[key.length] ; 
      int k = 0 ; 
      for (int i = 0 ; i < key.length ; i++) {
         if (key[i]) { ord[k] = i ; k++ ; } 
      }
      for (int i = 0 ; i < key.length ; i++) {
         if (!key[i]) {ord[k] = i ; k++ ; } 
      }
      return ord ;
   }

   /**
     *  @return whether boolean[] key is factual (all true)
      */
   protected static boolean factual(boolean[] key) {
      for (int i = 0 ; i < key.length ; i++) 
         if (!key[i]) return false ;
      return true ; 
  }

   /**
     *  Determine match order for Rewrite
     *  for Lterm in Rewrite A -> B.
     *  E.g., in X+0=X,
     *    Lterm = X + 0.  Lterm.order = [1,0]
     *  For f(x1,...,xn) = R, only set match-order
      *   for f.args, and not for the args in turn 
      *   (at this time).
      */
   protected static void setOrder(Rewrite m) {
      if (m.Lterm instanceof Variable) return ;
      // otherwise ... 
      Functor f = (Functor)(m.Lterm) ;   
      //System.out.println(f+":"+f.key) ;
      f.order = order(f.key) ;
      f.factual = factual(f.key) ;
   }

///////////////////////////////
///// END  setOrder 
///////////////////////////////
///////////////////////////////
///// BEGIN  setKeyPatterns    
///////////////////////////////
   /**
     *  For Rules and static Rewrites 
     *  and for all functors Inferences,
     *  record key patterns. Also,
     *  set theory.keyPatterns (for rule literals),
     *  and set theory.rewriteKeyPatterns
     *  Functor indexing called first.  
      */
   protected static void setKeyPatterns(Theory theory) {
      PatternVector[] keyPatterns = new PatternVector[theory.name.length] ;    // * generic array trick
      PatternVector[] rewriteKeyPatterns = new PatternVector[theory.name.length] ; // NEW   * generic array trick
      Iterator infs = theory.iterator() ; 
      while (infs.hasNext()) { 
         Inference inf = (Inference)infs.next() ; 
         if (inf instanceof Rule) { 
            Rule r = (Rule)inf ;
            Iterator ants = ((Rule)inf).antecedent.iterator() ;
            while (ants.hasNext()) { 
               Functor f = (Functor)ants.next() ; 
               Pattern pat = new Pattern(f.key) ;  
               if (keyPatterns[f.index] ==null) keyPatterns[f.index] = new PatternVector() ;
               if (!keyPatterns[f.index].contains(pat)) keyPatterns[f.index].add(pat) ; 
            }
            Iterator cons = ((Rule)inf).consequent.iterator() ; 
            while (cons.hasNext()) {
              Conjunction c = (Conjunction)cons.next() ; 
              Iterator cit = c.iterator() ;
              while (cit.hasNext()) { 
                 Functor f = (Functor)cit.next() ; 
                 Pattern pat = new Pattern(f.key) ; 
                 if (keyPatterns[f.index] ==null) keyPatterns[f.index] = new PatternVector() ;
                 if (!keyPatterns[f.index].contains(pat)) keyPatterns[f.index].add(pat) ; 
              }
            }
         }
         else if (inf instanceof Rewrite) { 
            Rewrite m = (Rewrite)inf ;
            if (m.Lterm instanceof Functor) {   
               Functor f = (Functor)m.Lterm ; 
               Pattern pat = new Pattern(f.key) ; 
               if (rewriteKeyPatterns[f.index] ==null) rewriteKeyPatterns[f.index] = new PatternVector() ;
               if (!rewriteKeyPatterns[f.index].contains(pat)) rewriteKeyPatterns[f.index].add(pat) ; 
            } 
         }
         else {} // no other cases at present
      }
      // trim pattern vectors to size here
      for(int i = 0 ; i < keyPatterns.length ; i++)
         if(keyPatterns[i] != null)  keyPatterns[i].trimToSize() ;
      theory.keyPatterns = keyPatterns ;  
      theory.rewriteKeyPatterns = rewriteKeyPatterns ;  
   }

///////////////////////////////
///// END  setKeyPatterns 
///////////////////////////////
////////////////////////////////////////////
///   BEGIN COMPLEXITY calculation methods
////////////////////////////////////////////
   // Complexity of disjunction is max complexity of disjuncts 
   protected static int complexity(Disjunction disj) {
      int c = 0 ;
      for(int i = 0 ; i < disj.size() ; i++) {
         int k =  complexity(disj.elementAt(i) ) ; 
         if (k > c) c = k ; 
      }
      return c ; 
   }

   // Complexity of conjunction is max complexity of conjuncts
   protected static int complexity(Conjunction conj) {
      int c = 0 ;
      for(int i = 0 ; i < conj.size() ; i++) {
         int k =  complexity(conj.elementAt(i) ) ; 
         if (k > c) c = k ; 
      }
      return c ; 
   }

   // Complexity of functor is max complexity of its terms 
   protected static int complexity(Functor f) {
      int c = 0 ; 
      for (int j=0 ; j < f.arity ; j++) {
         int k =  termComplexity(f.args[j]) ; 
         if (k > c) c = k ;
      }
      return c;
   }

   // static term complexity 
   protected static int termComplexity(Term t) {
      if (t instanceof Variable)  return 0 ; 
      else if (((Functor)t).arity == 0) return 0 ; 
      else { 
         int c = 0 ; 
         for (int j=0 ; j < ((Functor)t).args.length ; j++) {
            // total number of operations 
            // p(g(f(a,b)))  =>  p(f(g(a),g(b))) is complex
            //c += termComplexity(((Functor)t).args[j]) ; 
            // alternate is DEPTH of operations ..
            // p(g(f(a,b)))  =>  p(f(g(a),g(b))) is NOT complex
            int k =  termComplexity(((Functor)t).args[j])  ; 
            if (k > c) c = k ;
         }
         return 1+c ; 
      }
   }

   /*** Is rule complex? */
   public static boolean complex(Rule r) { 
      return complexity(r.antecedent) < complexity(r.consequent) ;
   }
//////////////////////////////
// end complexity calculations
//////////////////////////////

   /**
     * @return Theory analysis report 
      */
   public static String report(Theory theory) {
      StringBuffer bf = new StringBuffer() ;        
      bf.append("AutoLogAnalyzer version 6/17/2023 \n") ; 
      bf.append("Functors by index:name"+NAS+"arity  \n") ; 
      bf.append(inspectNames(theory)) ; 
      //bf.append("theory.subfactual="+theory.subfactual) ;  // removed from Theory
      // ADD comment re .index vs parametric indexical
      bf.append("[These index:name/arity values are based on SOURCE input syntax only, \n") ;  
      bf.append("and NOT on any subsequent indexicality/relevance calculations.] \n") ;  
      ///// 
      bf.append("\nAutoLog Inference attributes\n") ; 
      Iterator thy = theory.iterator() ;
      int n = 1 ; // number the Inferences, 1-based
      while(thy.hasNext()) {
         Inference inf = (Inference)thy.next() ;
         if (inf instanceof Rule) bf.append(n+". "+inspect((Rule)inf)) ; 
         else if (inf instanceof Rewrite) bf.append(n+". "+inspect((Rewrite)inf)) ;
         n++ ;
      } 
      bf.append("\n") ;    ///////////////
      /////    
      //code A moved to END NOTES
      /////
      bf.append(theory.showTheorySummary()) ;
      bf.append("\nSee info pane ??? for explanation of parameters and methods.\n") ;


      return bf.toString() ; 
   } // end report

   // Show functor names fields
   private static String inspectNames(Theory theory) { 
      StringBuffer sb = new StringBuffer() ; 
      for (int i = 0 ; i < theory.name.length ; i++) { 
         sb.append(i+" : " + theory.name[i] + NAS + theory.arity[i] + "   ") ; 
         String litpats = string(theory.keyPatterns[i]) ;
         String rewritepats = string(theory.rewriteKeyPatterns[i]) ;
         if (!litpats.isBlank())  
             sb.append(" rule-" + litpats) ; 
         if (!rewritepats.isBlank())  
             sb.append(" rewrite-" + rewritepats) ; 
         sb.append("\n") ; 
      }
      return sb.toString() ;
   }

   // rules have fixed data so trim Vectors to size
   private static void trimToSize(Theory theory) { 
      theory.trimToSize() ; 
      Iterator rit = theory.iterator() ;
      while (rit.hasNext()) {  
         Inference inf = (Inference)rit.next() ;
         if (inf instanceof Rule) { 
            Rule r = (Rule)inf ; 
            r.antecedent.trimToSize() ; 
            r.consequent.trimToSize() ; 
            Iterator cit = r.consequent.iterator() ;  
            while (cit.hasNext()) ((Conjunction)cit.next()).trimToSize() ; 
         }
      }
   }

   // string shows modulator data
   private static StringBuffer inspect(Rewrite m) { 
      String spc = " ", nl = "\n" ;
      StringBuffer sb = new StringBuffer() ; 
      sb.append(m.toString()) ; sb.append(nl) ;
      sb.append("  Rewrite data:  "+"U="+m.U+" V="+m.V + nl) ; 
      sb.append("             "+"Lterm = "+m.Lterm +" index="+m.Lterm.index + nl) ;
      sb.append("             "+"Rterm = "+m.Rterm +" index="+m.Rterm.index + nl) ;
      sb.append("             "+m.varindex+" {Var:index}" + nl) ; 
      if (m.converse != null) {
         sb.append("             converse = #"+ (m.converse.index +1) + nl) ; 
      } 
      if (m.Lterm instanceof Functor && ((Functor)(m.Lterm)).arity>0) { 
         sb.append("             Lterm.key=  "+string(((Functor)(m.Lterm)).key)+ nl) ;
         sb.append("             Lterm.order="+string(((Functor)(m.Lterm)).order)+ nl) ; 
      } 
      sb.append("             complex=" + m.complex + " parapredicative(Lterm)="+m.parapredicative + " factual="+m.factual+nl) ;
      if (!(m.warnings=="")) sb.append("   " + m.warnings + nl) ;
      return sb ; 
   } 

   // string shows rule fields
   private static StringBuffer inspect(Rule r) { 
      String spc = " ", nl = "\n" ;
      StringBuffer sb = new StringBuffer() ; 
      sb.append(r.toString()) ; sb.append(nl) ;
      sb.append("  Rule data: A="+r.A+" C="+r.C+" U="+r.U+" V="+r.V + " existential="+r.existential+ " predicative="+r.isPredicative()+ " indexical="+r.isIndexical() +"\n") ; 
      sb.append("             "+r.varindex+" {Var:index}\n") ; 
      sb.append("             factual="+r.factual+" complex="+r.complex+"\n") ; 
      Conjunction c = r.antecedent ;
      sb.append("   antecedent literal factor data: \n") ; 
      for (int i = 0 ; i < c.size() ; i++) {
         Functor f = (Functor)c.elementAt(i) ; 
         sb.append(spc+spc+spc+i+spc+"index="+f.index+" key="+string(f.key)+" order="+string(f.order)+" forward="+r.forward[i].toString()+nl) ; 
         // sb.append(spc+spc+spc+spc+spc+"predicative="+f.predicative+" indexical="+f.indexical+" factual="+f.factual+nl) ;  // add deep
         sb.append(spc+spc+spc+spc+spc+"predicative="+f.predicative+" .indexical="+f.indexical+" .isIndexical()="+f.isIndexical()+" factual="+f.factual+nl) ; 

      }
      sb.append("      =>") ; sb.append(nl) ;
      sb.append("   consequent literal factor data: \n") ; 
      Disjunction d = r.consequent ;
      Iterator dit = d.iterator() ;
      int ditcount = 0 ;
      while (dit.hasNext()) { 
         c = (Conjunction)dit.next() ;
         sb.append("      "+ditcount+" <conjunction>.factual="+c.factual+nl) ;
         for (int i = 0 ; i < c.size() ; i++) {
            Functor f = (Functor)c.elementAt(i) ; 
            sb.append("         "+i+" index="+f.index+" key="+string(f.key)+" order="+string(f.order)+nl) ; 
            sb.append("         "+"predicative="+f.predicative+" indexical="+f.indexical+" .isIndexical()="+f.isIndexical()+" factual="+f.factual+nl) ; // deep
         }
         if (dit.hasNext()) { 
            sb.append("      |"+nl) ; 
         }
         ditcount++ ;
      }
      if (!r.isIndexical())
      sb.append("   WARNING: Rule is NOT indexical. \n") ; 

      return sb ; 
      
   }

   // string shows predicate key
   protected static String string(boolean[] k) { 
      StringBuffer sb = new StringBuffer() ; 
      sb.append("[") ; 
      for(int i = 0 ; i < k.length ; i++) {
         sb.append(k[i]) ; 
         if (i < k.length -1) sb.append(",") ; 
     }
      sb.append("]") ; 
      return sb.toString() ; 
   }

   // string shows predicate order
   protected static String string(int[] a) { 
      StringBuffer sb = new StringBuffer() ; 
      sb.append("[") ; 
      for(int i = 0 ; i < a.length ; i++) {
         sb.append(a[i]) ; 
         if (i < a.length -1) sb.append(",") ; 
      }
      sb.append("]") ; 
      return sb.toString() ; 
   }

   // string shows PatternVector
   protected static String string(Vector<Pattern> v) { 
      if (v == null) return " " ;
      StringBuffer sb = new StringBuffer() ;  
      sb.append("keys=<") ; 
      for(int i = 0 ; i < v.size() ; i++) {
         sb.append(string(((Pattern)v.elementAt(i)).key)) ;
         //sb.append("\n") ; 
      }
      sb.append(">") ; 
      return sb.toString() ; 
   }

}  // end AutoLogAnalyzer

/* END NOTES
Code A
      bf.append("\ntheory.isIndexical()="+theory.isIndexical()+ "\n") ;
      // indexicality checker  
      n = 1 ; // reset inference number
      Iterator check = theory.iterator() ;
      while(check.hasNext()) {
         Inference inf = (Inference)check.next() ;
         bf.append("   ") ;
         if (inf instanceof Rule)  
            bf.append(n+".isIndexical()="+((Rule)inf).isIndexical()) ; 
         else if (inf instanceof Rewrite) 
            bf.append(n+".isIndexical()="+((Rewrite)inf).isIndexical()) ; 
         bf.append("\n") ;
         n++ ;
      } 

*/
