package sam.machine.lang ;  
import java.util.* ; 
import sam.util.* ; 

/**
  *  @version 5/12/2022 
  *  Rewrite .isIndexical() ≈ paraPredicative(Lterm) & .existential==0
  *  AutoLog static Rewrite modulator
  *  Source input program code form:    
  *    "Lterm = Rterm."  Left and Right terms
  *  ActiveRewrite Lterm→Rterm     
  *        replace match for Lterm in a branch fact 
  *        with corresponding instance of Rterm. 
  *  AutoLogAnalyzer sets converses (11/20/2021), see end comments. 
   */
public class Rewrite extends Inference  { 
   // set by constructor
   public Term Lterm ;        
   public Term Rterm ;         
   // coverses via AutoLogAnalyzer
   //        L = R
   //         \ /
   //          .
   //         / \
   //        R   L
   // public boolean primary ;   // is 1st specified ???
   public Rewrite converse ;        
   public int converse_index = -1 ; 
   /* 
      When Lterm is Functor, Lterm will have .key and .order set by 
      AutoLogAnalyzer
         Lterm boolean[] key -- key pattern for Lterm args, bound or not
         Lterm int[] order  --   satisfy Lterm in this order
   */
   public boolean predicative ;  // set by AA, OLD, only checks Lterm  
                                 // use -.isPredicative() instead
   public boolean parapredicative ; // requirement for Lterm, set by constructor
   public boolean factual ;   // Lterm bound 
   public int existential ;   // e.g., e = a*B. free variable in Rterm 
                              // = V-U, set by AutoLogAnalyzer
   public int U ;             // number of variables in Lterm
   public int V ;             // number of variables in rewrite
   public boolean complex ;   // Rterm more complex than Lterm
   public String varindex ;   // variable index for the entire rewrite
   public String warnings ;   // if not indexical, set by AA

   public Rewrite(Term L, Term R) { 
      this.Lterm = L ; 
      this.Rterm = R ; 
      this.parapredicative = (this.Lterm instanceof Functor) && 
                                   paraPredicative(this.Lterm) ; // X=.. not parapredicative
      this.warnings = "" ;  // can be modified by analyzer 
      if (!this.parapredicative) warnings = "WARNING: Lterm not parapredicative. " ; 
   }

   // public Rewrite(Functor equality) { // TBD, dynamic construction ? } 


   /**  
     *  NEW DEFINITION 4/13/2022
     *  A static Rewrite isIndexical() provided that
     *  the Lterm is paraPredicative and
     *  and that the Rterm has no free variable.
     *  E.G.,     apply(P)(A) = P(A).  
     *  This requires casting -- experimental & (nontrivial)
     *  and widens scope of rewrite indexicality. 
     *  Previously indexical rewrites are still indexical.
      */
   public boolean isIndexical() {
      if ( this.parapredicative && 
              this.existential==0 )   return true ; 
      return false ;
   }     

     /* -------------------------------------------------
            PREVIOUS SPECIFICATION 
            A Rewrite r is isIndexical() provided that
            its Lterm is a Functor and that both its
            Lterm and and Rterm are bounded parametric,
            and that the Rterm has no free variable.
           
        public boolean isIndexical() {
           if( this.Lterm instanceof Functor && 
                 this.Lterm.isPredicative()    &&
                   this.Rterm.isPredicative()  &&
                     this.existential==0)   return true ; 
           return false ;
        }     
     -----------------------------------------------  */

   /**
     *  static rewrites are VERY CONSTRAINED 
     *  -.isPredicative() ≡ -.isIndexical()
      */
   public boolean isPredicative() {
      if( this.Lterm instanceof Functor && 
            this.Lterm.isPredicative()    &&
              this.Rterm.isPredicative()  &&
                this.existential==0)   return true ; 
      return false ;
   }     

   /**
     *  byName():   isPredicative() and
     *  neither Lterm nor Rterm is parametric
      */
   public boolean byName() {
      if ( this.isPredicative() &&
             this.Lterm.byName() &&
               this.Rterm.byName() ) return true ; 
      return false ;
   }     

   /**
     * Display as string.
     */
   public String toString() { 
      StringBuffer sb = new StringBuffer() ; 
      sb.append(this.Lterm.toString()) ;
      sb.append(" = ") ; 
      sb.append(this.Rterm.toString()) ;
      sb.append(".") ; 
      return sb.toString() ; 
   }

   /**
     * Is R a converse rewrite to this? 
     *  e.g., 
     *     a(X,Y)=b(Y,X).   
     *     b(Y,X)=a(X,Y).  
     *
     *  a(X,Y)  b(Y,X)
     *       \ / 
     *        =       // MUST APPEAR CROSS-IDENTICAL
     *       / \      // (for now)
     *  a(X,Y)  b(Y,X)
     * 
     * AutoLogAnalyzer sets .converse for these converse rewrites
      */
   public boolean converse(Rewrite r) { 
      return (this.Lterm.toString().equals(r.Rterm.toString()) &&
               this.Rterm.toString().equals(r.Lterm.toString())) ;
   }

   /**
     *  A Term is para-predicative 
     *  provided that it is a Functor that is 
     *  either named, or the leftmost parametric .fnctr prefix is named.
     *  For example, f(g(A))(g(B))(U,V) is parapredicative.
     *  Any  prefix functor's name provides a context indexing
     *  the rest of the Functor. The method has "tricky" code.
     *  AND AN ERROR, use example F(X)(Y)=F(X,Y) to see...
      */
   public static boolean paraPredicative(Term t) {  
      if (t instanceof Functor) {
         Functor f = (Functor)t ; 
         if (f.fnctr != null)  {  // f is parametric
            if (f.fnctr instanceof Variable) return false ;
            else if (!paraPredicative(f.fnctr)) return false ;
            else { // f.fnctr is paraPredicative
               Functor ff = (Functor)f.fnctr ;             
               for (int i=0; i<ff.arity ; i++)  
                 if (!paraPredicative(ff.args[i])) return false ;
            }
         }  
         else { // f is named, so check args 
            // for (int i=0; i<f.args.length ; i++)  
            for (int i=0; i<f.arity ; i++)  
               if (!paraPredicative(f.args[i])) return false ;
         } 
      }
      else  return true ; // internal arguments
      return true ;
   }

   /**
     *  Test stubs for a translated small theory
     *  Call to analyzer included.
      */
   public static void main(String[] args) { 
      // "semipredicative" tests
        Theory theory = Theory.translate("X=f(X).") ;
//      Theory theory = Theory.translate("a = b.") ; 
//      Theory theory = Theory.translate("g(X)(Y)=f(X,Y).") ;
//      Theory theory = Theory.translate("X=f(X).") ;
//      Theory theory = Theory.translate("F(X)=X.") ;
//      Theory theory = Theory.translate("a=f(a).") ; // was cast error(fixed)
      AutoLogAnalyzer.analyze(theory) ;
      Rewrite r = (Rewrite)theory.elementAt(0) ;
      System.out.println("r is "+r) ;
      System.out.println("Lterm="+r.Lterm) ;
      System.out.println("Lterm.parapredicative? "+r.parapredicative) ;
      //System.out.println("Lterm parapredicative? "+Rewrite.paraPredicative(r.Lterm)) ;
      System.out.println("r byName? "+r.byName()) ;
      System.out.println("r existential= "+r.existential) ;  
      System.out.println("r isIndexical()="+r.isIndexical()) ;
      System.out.println("Lterm isPredicative()="+(r.Lterm).isPredicative()) ;  
      System.out.println("Rterm isPredicative()="+(r.Rterm).isPredicative()) ;  /// ???
      //
      //  converses test, see endnote 1-
   } // end main

} // end Rewrite


/* ----------------------------------------------------------------------
paraPredicative tests, e.g.,
     f(A)(B)(U,V)=g(U,V,A,B).
     h(f(X))=g(f(X)).
     X=f(X).
     a=f(a).
     a=f(Z).

1-      {cross converse test/check in main.  
           preliminary, not yet followed up as of 4/12/2022. }
      Theory theory = Theory.translate("a(X,Y)=b(Y,X).  b(Y,X)=a(X,Y).") ;
      System.out.println(theory) ; // 
      Rewrite r1 = (Rewrite)theory.elementAt(0) ;
      Rewrite r2 = (Rewrite)theory.elementAt(1) ;
      System.out.println(r1.Lterm) ;
      System.out.println(r2.Rterm) ;
      System.out.println(r1.converse(r2)) ;
      System.out.println(r2.converse(r1)) ;

   cross converse must be adjacent adjacent rewrites
   which appear identical, for now.  We could allow 
   cross-unify and not adjacent in the source code.
   « Consider to design a term unifier in Febuary 2021.
     sam/gui/search/Analyzer.html should note "converse" rewrites. »
2- NB 1/9/2021:  Rewrite extends Inference, not Functor.
   ??????????? constructor design needed here:
               new Rewrite(Functor equality)
3- NB 4/13/2022: NOT decided whether Autolog22 will 
   "formally support" converses.
----------------------------------------------------------------------- */
