package sam.machine.lang ;  

import java.util.* ; 
import sam.util.* ; 

/**
  * AutoLog Rule  
  * @version 11/20/2021  
  *   
   */
public class Rule extends Inference {
   
   public String annotation ;    // later  (* ... *)
   public Conjunction antecedent ; 
   public Disjunction consequent ; 
   public int existential ;   // # free variables in consequent
   public int A,  // width of antecedent
              C,  // A + widest consequent
              U,  // number of universal variables in the antecedent, set by Analyzer
              V ; // total number of variables, set by AutoLogAnalyzer
   public boolean complex ;  // complex rule?  Set by AutoLogAnalyzer
   public boolean factual ;  // has factual antecedent?  Set by AutoLogAnalyzer
   //public boolean subfactual ;  // has subfactual antecedent?  deprecated
   public Index[] forward ;  // consequent factual checking
   public String varindex ;  // variable index report from AutoLogAnalyzer

   public Rule(Conjunction c, Disjunction d) { 
      this.antecedent = c ; 
      this.consequent = d ; 
      this.A = this.antecedent.size() ; 
      int w = 0 ; 
      Iterator it = this.consequent.iterator() ; 
      while (it.hasNext()) { 
         int ww = ((Conjunction)it.next()).size() ; 
         if (ww > w) w = ww ; 
      }
      this.C = this.A + w ;  
   }

   /**
     * Display as string.
     */
   public String toString() { 
      StringBuffer sb = new StringBuffer() ; 
      sb.append(this.antecedent.toString()) ;
      sb.append(" => ") ; 
      sb.append(this.consequent.toString()) ;
      sb.append(".") ; 
      return sb.toString() ; 
   }

   /**
     *  Whether all literals are predicative, 
     *  after AA calculates Functor.predicative.
     *  "Deep inspection" version.
      */   
   public boolean isPredicative() { 
      Iterator it = this.antecedent.iterator() ; 
      while (it.hasNext()) 
         if (!((Functor)(it.next())).isPredicative()) return false ;
      it = this.consequent.iterator() ;
      while (it.hasNext()) {
         Iterator cit = ((Conjunction)(it.next())).iterator() ;
         while (cit.hasNext()) 
            if (!((Functor)(cit.next())).isPredicative()) return false ; 
      }
      return true ; 
   }

   /**  
     *  Whether all functors inside Rule literals are indexical, 
     *  called AFTER AutoLogAnalyzer.  
     *  "Deep inspection" version.
      */   
   public boolean isIndexical() {
      Iterator it = this.antecedent.iterator() ; 
      while (it.hasNext()) 
         if (!((Term)(it.next())).isIndexical()) return false ;     // N.B.
      it = this.consequent.iterator() ;
      while (it.hasNext()) {
         Iterator cit = ((Conjunction)(it.next())).iterator() ;
         while (cit.hasNext()) 
            if (!((Term)(cit.next())).isIndexical()) return false ; // N.B.
      }
      return true ; 
   }
    
   /**  
     *  Earlier shallow version intended for Rule literals only
     *  Whether all to-level Rule literals are indexical, 
     *  called AFTER AutoLogAnalyzer.  
     *  "SHALLOW indexicality" for Rule literal. 
      *  NOT CALLED.
      */   
   public boolean isLiteralIndexical() {
      Iterator it = this.antecedent.iterator() ; 
      while (it.hasNext()) 
         if (!((Functor)(it.next())).indexical) return false ;
      it = this.consequent.iterator() ;
      while (it.hasNext()) {
         Iterator cit = ((Conjunction)(it.next())).iterator() ;
         while (cit.hasNext()) 
            if (!((Functor)(cit.next())).indexical) return false ; 
      }
      return true ; 
   }


   /** 
     *  whether all literals in this rule are byName() 
     *  Each literal factor LF has LF.fnctr==null
     */
   public boolean byName() { 
      Iterator ait = this.antecedent.iterator() ;
      while (ait.hasNext()) 
         if ( !((Functor)ait.next()).byName() ) return false ;
      Iterator dit = this.consequent.iterator() ; 
      while (dit.hasNext()) { 
         Iterator cit = ((Conjunction)dit.next()).iterator() ;
         while (cit.hasNext()) 
            if ( !((Functor)cit.next()).byName() ) return false ;
      }
      return true ;
   }

} // end Rule



