package sam.machine.lang ;

/**
  *  AutoLog Term
  *    hierarchy 
  *      Variable < Term
  *      Functor < Term
  *      Literal < Functor < Term  
  *         { Literal inhabits Rule }
  *      Fact < Term        
  *         { Fact inhabits Branch }
  *  version 11/21/2021
   */
public abstract class Term {
   public String name ;
   public int index ; 
   public int arity ;

   /**
     *  Generate string representation of this Term 
      */
   public abstract String toString() ;

   /**
     *  Is this term deep indexical?
     *  Called after AutoLogAnalyzer analyzes 
     *  input program. 
      */
   public boolean isIndexical() { 
      if (this instanceof Functor) { 
         Functor f = (Functor)this ;
         if (! f.indexical) return false ; // marked by analyzer
         for (int i=0 ; i < this.arity ; i++)       // deeper
            if (! f.args[i].isIndexical()) return false ;
         return true ;  // indexical functor with indexical args
      }
      else return true ; // Variable arg is indexical by default
   }

   /**
     *  Is this term deep predicative?
     *  Called after AutoLogAnalyzer analyzes 
     *  input program. 
     *  @version 11/13/2021
      */
   public boolean isPredicative() { 
      if (this instanceof Functor) { 
         Functor f = (Functor)this ;
         if (! f.predicative) return false ; // marked by analyzer
         for (int i=0 ; i < this.arity ; i++)       // deeper
            if (! f.args[i].isPredicative()) return false ;
         return true ;  // predicative functor with predicative args
      }
      else return true ; // true for Variable arg by default
   }

   /**
     *  Is this term byName(), 
     *  No parametricity.
     *  @version 11/17/2021
      */
   public boolean byName() { 
      if (this instanceof Functor) { 
         Functor f = (Functor)this ;
         if (!(f.fnctr==null)) return false ;  // named 
         for (int i=0 ; i < this.arity ; i++) // arguments also
            if (! f.args[i].byName()) return false ;
         return true ;  // named functor with named args
      }
      else return true ; // true for Variable arg by default
   }

   /**
     *  test stub 
      */
   public static void main(String[] args) {
   }

}

